
//
///////////////////////////////////////////////////////////////////////////////
//////   This file is created by a script. Modifications will be lost!   //////
///////////////////////////////////////////////////////////////////////////////

#ifndef SZS_DB_OBJECT_H
#define SZS_DB_OBJECT_H 1

#include "lib-std.h"

//
///////////////////////////////////////////////////////////////////////////////
///////////////////////////   enum ObjectFlags_t   ////////////////////////////
///////////////////////////////////////////////////////////////////////////////

typedef enum ObjectFlags_t
{
    OBF_REQUIRED     = 0x00000001,  // for track files required
    OBF_REQUIRED2    = 0x00000002,  // alternativley required
    OBF_OPTIONAL     = 0x00000004,  // optional file
    OBF_OBJECT       = 0x00000008,  // file is object relevant
    OBF_EXTERNAL     = 0x00000010,  // external file
    OBF_UNIQUE       = 0x00000020,  // all instances are identical
    OBF_BRRES        = 0x00000040,  // B: at least one BRRES file exists
    OBF_EFFECT       = 0x00000080,  // E: at least one effect pair (BREFF+BREFT) exists
    OBF_BRASD        = 0x00000100,  // D: at least one BRASD file exists
    OBF_KCL          = 0x00000200,  // K: at least one KCL file exists
    OBF_KMP          = 0x00000400,  // P: at least one KMP file exists
    OBF_OTHER        = 0x00000800,  // x: at least one unknown file exists
    OBF_NO_FILE      = 0x00001000,  // ?: no file found for group
    OBF_SCALE        = 0x00002000,  // is known as scalable
    OBF_ROTATE       = 0x00004000,  // is known as rotatable
    OBF_FALLDOWN     = 0x00008000,  // relevant for fall-down
    OBF_OBJFLOW      = 0x00010000,  // defined in ObjFlow.bin
    OBF_SOLID        = 0x00020000,  // object is solid
    OBF_ROAD         = 0x00040000,  // object acts like a road
    OBF_NO_ROUTE     = 0x00080000,  // object needs no route
    OBF_OPT_ROUTE    = 0x00100000,  // object has an optional route
    OBF_ALWAYS_ROUTE = 0x00200000,  // object has always a route
    OBF_FNAME        = 0x00400000,  // object need at least 1 file
    OBF_MULTI_FNAME  = 0x00800000,  // object needs 2 or more files
    OBF_F_TRACK_S    = 0x01000000,  // object found in standard track file(s)
    OBF_F_TRACK_D    = 0x02000000,  // object found in _d track file(s)
    OBF_F_ARENA_S    = 0x04000000,  // object found in standard arena file(s)
    OBF_F_ARENA_D    = 0x08000000,  // object found in _d arena file(s)
    OBF_F_OTHER      = 0x10000000,  // object found in other file(s)
    OBF_ITEMBOX      = 0x20000000,  // object is an itembox

    OBF_M_ROUTE      = OBF_NO_ROUTE | OBF_OPT_ROUTE | OBF_ALWAYS_ROUTE,
    OBF_M_TRACK      = OBF_F_TRACK_S | OBF_F_TRACK_D,
    OBF_M_ARENA      = OBF_F_ARENA_S | OBF_F_ARENA_D,
    OBF_M_FOUND      = OBF_M_TRACK | OBF_M_ARENA | OBF_F_OTHER,

    OBF_NONE         = 0x00000000,  // none set
    OBF_ALL          = 0x3fffffff   // all flags of above

} ObjectFlags_t;

//
///////////////////////////////////////////////////////////////////////////////
///////////////////////////   struct ObjectInfo_t   ///////////////////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct ObjectInfo_t
{
    ccp			name;		// NULL or name of object
    ccp			info;		// NULL or info for object
    ObjectFlags_t	flags;		// flags of object
    ccp			charact;	// characteristics based on 'flags'
    ccp			format1;	// format for first 4 settings
    ccp			format2;	// format for second 4 settings
    int			group_idx[3];	// >=0: index into 'DbFileGROUP'
    ccp			fname;		// NULL or filename list
    ccp			param[8];	// NULL or parameter info

} ObjectInfo_t;

#define N_KMP_GOBJ 756
extern const ObjectInfo_t ObjectInfo[N_KMP_GOBJ];


//
///////////////////////////////////////////////////////////////////////////////
/////////////////////////////   format strings   //////////////////////////////
///////////////////////////////////////////////////////////////////////////////

extern const char settings_format_uuuu[];
extern const char settings_format_uxuu[];
extern const char settings_format_uxxu[];

//
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////   statistics   ////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#define OBJ_FW_NAME     18
#define OBJ_FW_INFO    753
#define OBJ_FW_FNAME    46
#define OBJ_FW_CHAR     55
#define OBJ_FW_SETTING 350


//
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////   E N D   //////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#endif // SZS_DB_OBJECT_H
