
/***************************************************************************
 *                         _______ _______ _______                         *
 *                        |  ___  |____   |  ___  |                        *
 *                        | |   |_|    / /| |   |_|                        *
 *                        | |_____    / / | |_____                         *
 *                        |_____  |  / /  |_____  |                        *
 *                         _    | | / /    _    | |                        *
 *                        | |___| |/ /____| |___| |                        *
 *                        |_______|_______|_______|                        *
 *                                                                         *
 *                             Wiimms SZS Tools                            *
 *                           http://szs.wiimm.de/                          *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This file is part of the SZS project.                                 *
 *   Visit http://szs.wiimm.de/ for project details and sources.           *
 *                                                                         *
 *   Copyright (c) 2011-2016 by Dirk Clemens <wiimm@wiimm.de>              *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   See file gpl-2.0.txt or http://www.gnu.org/licenses/gpl-2.0.txt       *
 *                                                                         *
 ***************************************************************************/

#include "dclib-types.h"

//
///////////////////////////////////////////////////////////////////////////////

static const u8 ctcode_tex_header[] =
{
  0x54,0x45,0x58,0x30, 0x00,0x08,0x76,0x40, 0x00,0x00,0x00,0x03, 0x00,0x00,0x00,0x00,
  0x00,0x06,0x58,0xc0, 0x00,0x08,0x76,0x44, 0x00,0x00,0x00,0x00, 0x02,0x60,0x01,0xc8,
  0x00,0x00,0x00,0x0e, 0x00,0x00,0x00,0x01, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,
};

//
///////////////////////////////////////////////////////////////////////////////

static const u8 ctcode_header[] =
{
  0xba,0xd1,0xda,0x7a, 0x00,0x01,0xbf,0xb4, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x05,
  0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x50, 0x43,0x55,0x50,0x31, 0x00,0x00,0x38,0x90, 0x43,0x52,0x53,0x31,
  0x00,0x01,0x42,0xd0, 0x4d,0x4f,0x44,0x31, 0x00,0x01,0x51,0xd0, 0x4d,0x4f,0x44,0x32,
  0x00,0x01,0x6c,0x50, 0x4f,0x56,0x52,0x31, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,
  // 0x50 = 80 Bytes
};

//
///////////////////////////////////////////////////////////////////////////////

static const u8 cup1_header[] =
{
  0x43,0x55,0x50,0x31, 0x00,0x00,0x38,0x40, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x36,
  0x00,0x00,0x00,0x02, 0x00,0x00,0x08,0x00, 0x00,0x00,0x08,0x00, 0x00,0x00,0x08,0x00,
  0x00,0x00,0x08,0x00, 0x00,0x00,0x08,0x00, 0x00,0x00,0x08,0x00, 0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,
  // 0x40 = 64 Bytes
};

//
///////////////////////////////////////////////////////////////////////////////

static const u8 crs1_header[] =
{
  0x43,0x52,0x53,0x31, 0x00,0x01,0x0b,0x40, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0xfa,
  0x00,0x00,0x00,0x2a, 0x00,0x00,0x08,0x00, 0x00,0x00,0x08,0x00, 0x00,0x00,0x08,0x00,
  0x00,0x00,0x08,0x00, 0x00,0x00,0x08,0x00, 0x00,0x00,0x08,0x00, 0x00,0x00,0x08,0x00,
  0x00,0x00,0x08,0x00, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00, 0x00,0x00,0x00,0x00,
  // 0x40 = 64 Bytes
};

//
///////////////////////////////////////////////////////////////////////////////

static const u8 ctcode_image_bz2[] =
{
  0x00,0x02,0x1d,0x90, 0x42,0x5a,0x68,0x32, 0x31,0x41,0x59,0x26, 0x53,0x59,0x14,0xfc,
  0x47,0x10,0x00,0xb4, 0x04,0x7f,0xff,0xfa, 0x70,0x67,0x5f,0xf7, 0x70,0x60,0x2f,0xff,
  0xf4,0xf7,0x7f,0xf2, 0x70,0x18,0x0f,0xfb, 0x7c,0x67,0x7f,0xf7, 0x76,0x02,0x1f,0x83,
  0x01,0x01,0x1f,0x81, 0x80,0x80,0x8f,0xe0, 0x21,0x3e,0x3d,0x45, 0x29,0xf2,0x94,0x89,
  0xde,0xc7,0xd5,0x5b, 0x7c,0xee,0xbc,0xfb, 0xc9,0x4b,0xed,0x55, 0x51,0x47,0xd0,0x2b,
  0xef,0x9a,0xdc,0x31, 0x54,0x02,0x05,0x14, 0xd2,0x9e,0x7b,0xef, 0xb3,0xd8,0xb4,0xd7,
  0xdd,0x63,0xe7,0xdd, 0xcd,0xc7,0x89,0xdf, 0x4e,0xe5,0x9d,0xc3, 0x9d,0x2f,0xa6,0xd8,
  0x3d,0x7d,0xf7,0xde, 0x50,0xee,0xb3,0xc9, 0xef,0xae,0x6f,0x2f, 0x1d,0x7b,0x7b,0xae,
  0xdf,0x6a,0xd9,0xbe, 0x8f,0xa6,0xd7,0xcf, 0x1f,0x3e,0xf5,0xb6, 0xf0,0xd3,0x44,0x4d,
  0x00,0x13,0x26,0x40, 0x01,0x32,0x22,0x32, 0x4f,0x32,0x9e,0x84, 0xf4,0x09,0x01,0xe8,
  0xd4,0x69,0xb5,0x0d, 0xa4,0x1a,0x00,0x1a, 0x00,0x00,0x0f,0x50, 0x0d,0x06,0x82,0x53,
  0x41,0x04,0x26,0x82, 0x04,0xd2,0x61,0x14, 0x7e,0x4c,0xa9,0xfa, 0x69,0x27,0xa8,0x7a,
  0x27,0xa9,0xa3,0xd4, 0x0d,0x34,0x06,0x83, 0x6a,0x03,0xd4,0x00, 0x0d,0x1e,0x90,0x34,
  0x00,0xc9,0xea,0x0d, 0x34,0x00,0x94,0xc5, 0x11,0x08,0xd4,0xf4, 0x4c,0x53,0x41,0xa6,
  0x8f,0x53,0xd2,0x1e, 0x90,0xd0,0xd3,0x4d, 0x3d,0x4d,0x00,0x34, 0x36,0x93,0xd4,0x06,
  0x83,0x40,0x00,0x00, 0x1a,0x00,0x00,0x00, 0x02,0x14,0x92,0x21, 0xa5,0x4f,0xcd,0x54,
  // 0x100:
  0xf5,0x37,0xa5,0x3d, 0x47,0x94,0x7a,0x80, 0x1f,0xa4,0x80,0xd0, 0x64,0x00,0x1a,0x00,
  0x00,0xf5,0x01,0xa0, 0x00,0xd0,0x68,0x0c, 0x83,0x40,0x00,0xd0, 0x21,0x52,0x88,0x49,
  0xbd,0x22,0x7a,0x40, 0xd0,0xd3,0xf5,0x4d, 0x3d,0x4c,0x43,0x4d, 0x0d,0x0d,0xa8,0xd8,
  0x91,0x93,0x40,0x68, 0x64,0x00,0x34,0x0d, 0x03,0x40,0x00,0x01, 0xa0,0x06,0x81,0xa0,
  0x93,0x51,0x21,0x34, 0x13,0x23,0x42,0x68, 0x1a,0x49,0xe1,0x51, 0xea,0x7e,0xa9,0xed,
  0x29,0xea,0x1f,0xaa, 0x64,0x79,0x4c,0x80, 0xd1,0xa1,0xfa,0xa0, 0x34,0x0c,0x86,0x80,
  0x68,0x31,0x00,0x00, 0x00,0x01,0xec,0x02, 0x88,0xa0,0xb0,0x28, 0x8a,0x0b,0xf2,0x76,
  0x98,0xd7,0x21,0x00, 0x21,0x10,0x91,0x00, 0x61,0x10,0x81,0x45, 0x41,0x1f,0x14,0x51,
  0x09,0x10,0x15,0x51, 0xc6,0xa6,0x7c,0x39, 0x15,0x40,0x0a,0x15, 0x40,0xa4,0x14,0x02,
  0x84,0xa4,0x1f,0x47, 0x87,0xad,0xdf,0xdd, 0xb9,0x28,0xa4,0x40, 0xa4,0x29,0x5e,0x73,
  0xc8,0xcb,0x58,0xf4, 0x00,0xb2,0x28,0x34, 0x22,0xa7,0x04,0x05, 0x84,0x42,0x84,0x43,
  0x0a,0x8c,0x8a,0xaa, 0x26,0x69,0x01,0x50, 0x39,0xc8,0x14,0x41, 0x38,0x40,0x8a,0x80,
  0x9b,0x84,0x60,0x0a, 0x06,0x24,0x5a,0x11, 0xa4,0xa0,0x66,0x82, 0x28,0x2a,0x99,0xaa,
  0x98,0x21,0x11,0x88, 0x46,0xa9,0x51,0x10, 0xe0,0x42,0x1d,0x10, 0x2a,0x0e,0xe8,0x51,
  0xe2,0x42,0xa8,0x69, 0x40,0x89,0x40,0x28, 0xf1,0x65,0x5d,0xbc, 0x0a,0x1d,0xe4,0xaa,
  0x9a,0x32,0xa0,0x6d, 0x01,0x4a,0xd0,0x52, 0x6e,0x94,0x47,0x90, 0x84,0x13,0x20,0x4e,
  // 0x200:
  0x30,0x02,0x64,0x01, 0xe0,0xc1,0xdf,0x87, 0x96,0x10,0x76,0x8e, 0x32,0xea,0x04,0x28,
  0x57,0xde,0x7b,0xec, 0x53,0x34,0x29,0xa1, 0x20,0xb9,0x23,0x46, 0x04,0x39,0x12,0x71,
  0xe0,0xda,0xca,0x1a, 0x71,0xd1,0xcb,0x9a, 0x55,0xe1,0x31,0x29, 0xd2,0xc2,0x64,0x9b,
  0xb9,0xb1,0x17,0x7c, 0x8b,0xab,0x7c,0x85, 0xbf,0x17,0x44,0x6e, 0xdf,0x82,0xb9,0x20,
  0xd2,0x1a,0x9e,0x58, 0x07,0x53,0xaa,0xb7, 0xb6,0xa0,0x5a,0xe6, 0x8d,0x95,0x84,0xb6,
  0x92,0x06,0xe6,0x1c, 0xa2,0x75,0x3b,0xb6, 0x35,0x26,0x62,0x01, 0xab,0x24,0x1a,0xa0,
  0xe4,0x82,0x22,0x9d, 0x2c,0x26,0x3b,0x0a, 0xcb,0xa6,0xa1,0xb1, 0x90,0xe3,0x93,0xc2,
  0xae,0x92,0x69,0x13, 0x82,0x14,0xec,0x7b, 0x30,0xa6,0x4b,0xb1, 0x03,0x92,0xea,0x77,
  0xc2,0xf6,0xb5,0xad, 0x0a,0x21,0x8e,0x28, 0x1c,0x37,0x97,0x2a, 0x30,0xe7,0x66,0xa8,
  0x4c,0x87,0x09,0x59, 0x92,0x6b,0xb6,0x83, 0x06,0x33,0xb8,0x5f, 0xbb,0x16,0x43,0x9a,
  0x69,0x19,0xc8,0x4e, 0x8c,0xb0,0x58,0x6f, 0x64,0xc3,0x0b,0x20, 0xea,0x91,0x06,0x43,
  0x6b,0xb7,0x41,0xa0, 0x5c,0x21,0xd3,0x23, 0xb6,0xc4,0x07,0x2e, 0x1a,0x03,0x92,0x50,
  0xde,0xfd,0x90,0xd6, 0x8d,0x89,0x32,0x50, 0xf2,0x24,0xda,0x93, 0x57,0x44,0xcb,0x15,
  0x37,0x26,0x04,0x82, 0xbf,0x25,0x42,0x97, 0x2d,0x97,0x09,0x55, 0x40,0xa0,0x66,0xc2,
  0xd8,0x0a,0x5a,0x00, 0xb0,0x9a,0x25,0x72, 0xbb,0x92,0xd8,0x06, 0x69,0x69,0x3e,0x7b,
  0x68,0x43,0x7c,0xae, 0xd1,0xa8,0x50,0xc9, 0x0a,0x0d,0xd0,0x86, 0x89,0xa7,0x69,0x03,
  // 0x300:
  0x24,0x4e,0xa8,0xe5, 0xd6,0x6d,0x27,0x02, 0x35,0x6d,0x0a,0x39, 0x0f,0x24,0x3c,0x92,
  0x1a,0x95,0xad,0x49, 0xab,0x7e,0xd9,0xa9, 0x3b,0x33,0xb7,0xd1, 0x93,0x12,0x24,0x4d,
  0xa1,0xdd,0x34,0x6f, 0x84,0xf3,0x9b,0xf1, 0xe1,0x29,0xf2,0xad, 0x1a,0x73,0x18,0x73,
  0x48,0x1d,0x82,0x1a, 0xb0,0xa6,0x1d,0x2c, 0xd1,0x90,0x0d,0xd4, 0x09,0x52,0x1e,0x50,
  0x08,0x43,0xbc,0x02, 0x4e,0xd8,0x65,0x3e, 0xc7,0x66,0x5e,0x9b, 0xf1,0x19,0x72,0xd0,
  0xb4,0x2c,0x3f,0x17, 0xcf,0x47,0x8a,0xb7, 0xa6,0x69,0x8c,0x47, 0xf0,0xda,0xf0,0xd2,
  0x22,0x68,0x04,0x05, 0x44,0xc0,0x2a,0xe3, 0xbd,0x58,0xca,0x63, 0x84,0xac,0x15,0x26,
  0xa6,0x08,0x89,0xf8, 0xc3,0xeb,0x9c,0x2f, 0x1e,0x43,0xf4,0xde, 0xb0,0xdb,0xf1,0xb0,
  0x6e,0x17,0xa0,0x23, 0x3d,0x54,0x65,0xaa, 0xb8,0x27,0x7d,0xf5, 0x23,0x9e,0xef,0xf1,
  0x19,0xf1,0xe7,0x9f, 0x39,0x8c,0xe1,0x20, 0x07,0x7d,0x1f,0x4e, 0xfc,0xb6,0x6d,0x7b,
  0x50,0xdf,0xf1,0x7b, 0xfa,0x38,0xe4,0xdc, 0x72,0xcc,0xb8,0x73, 0xc5,0x13,0xe5,0x62,
  0x57,0xa5,0x37,0x77, 0x57,0x23,0x49,0xc1, 0x3d,0xfd,0x9e,0x7f, 0x27,0xa6,0x6c,0xcf,
  0xea,0x6b,0x9e,0x58, 0x3d,0x59,0x9d,0x0a, 0xf3,0x61,0x8e,0x10, 0x0a,0x64,0x02,0xab,
  0x63,0x70,0x05,0x02, 0x0a,0x49,0x16,0x41, 0x00,0x95,0x67,0x8a, 0xcb,0x60,0x5d,0x79,
  0x5c,0xae,0xb5,0xe0, 0xe6,0xa4,0x38,0x8e, 0x16,0x46,0x34,0x9d, 0xfa,0x55,0x60,0xa1,
  0x16,0xb2,0x8e,0x2d, 0x4b,0x42,0xd4,0x3d, 0x88,0xe8,0xb4,0xc8, 0x1d,0x8f,0xdb,0xe5,
  // 0x400:
  0x82,0xd8,0xb3,0x56, 0x6a,0xc9,0x87,0x0c, 0x94,0x67,0x55,0x74, 0x7b,0xaa,0xdc,0xd3,
  0x10,0x51,0xb0,0xd2, 0x6d,0x28,0xcc,0xf0, 0x75,0x99,0x21,0x0d, 0x3a,0x66,0x15,0x0a,
  0x74,0xae,0x34,0x36, 0xd2,0x28,0xc9,0x2f, 0x22,0xb1,0xc3,0xc9, 0x66,0x79,0x98,0x65,
  0xb5,0x04,0x78,0x56, 0xf6,0x5b,0x14,0xc8, 0x65,0x5d,0x1a,0x67, 0x49,0x0e,0x9e,0x54,
  0x0f,0x2e,0xcc,0xa9, 0x45,0xd9,0x6a,0x52, 0x25,0x58,0x04,0x19, 0x91,0x5a,0x34,0xae,
  0x31,0xda,0x44,0x45, 0x62,0x17,0xa9,0xfc, 0x25,0x9d,0xb4,0xb5, 0xe2,0x4b,0x8e,0xd8,
  0x63,0x2f,0x24,0x08, 0x81,0x22,0x9d,0xc9, 0x43,0x97,0x6d,0x96, 0xa8,0x22,0x52,0x4d,
  0x2e,0xb0,0x87,0x82, 0x64,0x61,0xba,0x42, 0xa7,0x2a,0xd1,0xb1, 0x9a,0x3f,0x1e,0xb8,
  0x2f,0x97,0x1c,0x6b, 0x20,0x08,0x07,0x37, 0x7a,0x3d,0xf9,0x54, 0xc5,0x7a,0x48,0xbc,
  0xdf,0x6d,0xaa,0x0a, 0xf9,0xf9,0xc4,0x51, 0x4c,0x5c,0xeb,0x11, 0x17,0xf2,0xa2,0xe6,
  0xcc,0x0a,0x22,0x82, 0xf4,0x33,0xc3,0xd0, 0xf2,0x83,0x9f,0xcd, 0x6c,0xb2,0x44,0xdd,
  0xaf,0x6d,0xdf,0xe7, 0xfc,0x71,0xbf,0xea, 0x0b,0xd7,0xcb,0x52, 0xee,0xb1,0x41,0x98,
  0x10,0x85,0x63,0x0f, 0x53,0x25,0x8f,0xf2, 0x95,0xda,0x9b,0x4c, 0xbd,0xa9,0xf0,0xbe,
  0x94,0x7d,0x7f,0x8d, 0x7d,0x9f,0x8a,0xe7, 0xca,0xb1,0xae,0x21, 0xbf,0xbc,0xd0,0xe9,
  0x45,0x73,0xe8,0xfd, 0x6c,0xc0,0xcd,0xf0, 0xa0,0xa2,0x83,0xeb, 0xfe,0x79,0xf5,0xa7,
  0x86,0x3a,0x0d,0xfc, 0xd2,0xd1,0x9d,0xb4, 0xb0,0x2c,0x3d,0x7e, 0x77,0x49,0x2b,0xd5,
  // 0x500:
  0xf7,0xbf,0x8e,0x7e, 0x43,0xf6,0x41,0xe8, 0xcb,0xd4,0x7a,0x5a, 0x32,0x5e,0xa1,0xf8,
  0x2a,0x0f,0x83,0xd0, 0x8d,0xfe,0x42,0xc2, 0xc8,0x3f,0x60,0x79, 0x7f,0xdf,0x2f,0x69,
  0x65,0x64,0x04,0x5f, 0x20,0x89,0x0d,0xcb, 0x8f,0xf5,0xc7,0x8b, 0x5e,0xe1,0xa2,0xb8,
  0xb3,0x07,0xe5,0x10, 0x54,0xe4,0x7d,0x7c, 0x28,0x32,0x9a,0x1d, 0x48,0x2f,0xc5,0xf0,
  0xe5,0xd2,0xe8,0x84, 0x43,0x3b,0x28,0x69, 0x16,0x32,0xed,0xaa, 0x21,0xdd,0x11,0x0f,
  0x27,0x00,0x17,0xb9, 0xd1,0x8a,0x88,0xfd, 0x09,0x14,0xdd,0x20, 0x01,0x08,0x84,0x0a,
  0x8a,0x25,0x02,0x06, 0xc4,0x80,0xaf,0x12, 0x76,0x24,0x00,0x0c, 0xb0,0x0a,0x65,0x93,
  0x15,0x3b,0x32,0xa1, 0xc5,0x80,0x4c,0x90, 0x29,0x40,0x22,0x14, 0x88,0x52,0xa8,0x03,
  0x4a,0xfc,0x08,0x45, 0x4c,0x80,0x0d,0xa5, 0x57,0x74,0x20,0x1a, 0x80,0x40,0x0a,0x04,
  0x14,0x0d,0x48,0x20, 0x29,0x48,0x26,0x4a, 0x00,0x1c,0x89,0x41, 0x15,0x71,0x08,0x29,
  0xcb,0x85,0x51,0x71, 0x0a,0x26,0xc4,0x83, 0xe0,0xc2,0x0e,0xa4, 0x10,0xf1,0x63,0x25,
  0x5a,0x5a,0x39,0x21, 0xc9,0x43,0xbf,0x1e, 0x34,0xae,0xa0,0x44, 0xe3,0x20,0xb8,0x48,
  0x85,0xf0,0xb0,0x0c, 0x8a,0x14,0xa1,0x57, 0x7c,0x8f,0x93,0x1a, 0x80,0xa6,0x94,0x4c,
  0xd6,0xa5,0xa7,0x20, 0x65,0x80,0xdc,0x49, 0x82,0x4d,0x28,0x31, 0x0d,0x0b,0xed,0x3c,
  0x9c,0x20,0x43,0x0c, 0xea,0x4a,0x5f,0xa1, 0x40,0x63,0xa2,0xa0, 0x61,0x02,0x16,0x85,
  0x27,0xd4,0x43,0xe7, 0xab,0xa9,0x4d,0xa3, 0xc6,0xb5,0x2e,0xa4, 0x72,0x47,0xa2,0x50,
  // 0x600:
  0xec,0x41,0xd3,0x3a, 0x85,0xa3,0xa7,0x7e, 0x1d,0x28,0x74,0x33, 0x62,0x06,0x6d,0xa1,
  0x8c,0x56,0x4c,0x50, 0x29,0x22,0x8a,0x51, 0xd3,0x41,0x69,0xf4, 0x6c,0x84,0xaa,0x22,
  0xa1,0x76,0xb7,0xc0, 0x6a,0x40,0x75,0x01, 0x92,0x14,0xf5,0x49, 0xbf,0x2a,0x99,0x33,
  0x24,0x8a,0x41,0x48, 0x1e,0x04,0x26,0xe7, 0xb5,0x90,0x32,0x60, 0x2c,0xe0,0x86,0xd4,
  0x3a,0x65,0x57,0x96, 0x0a,0x06,0x84,0x0a, 0xa1,0x29,0x57,0x74, 0x70,0x8d,0x42,0x1e,
  0x03,0x49,0xc1,0xa4, 0x8b,0x00,0x14,0x27, 0x17,0x54,0x03,0xa9, 0x09,0x3d,0xc9,0x21,
  0xda,0x84,0x3a,0x35, 0xc1,0x5e,0x78,0xe6, 0x82,0x25,0x38,0xc2, 0x70,0xb6,0x80,0x7a,
  0x24,0x37,0x4d,0x0d, 0x6b,0x30,0x3a,0x99, 0x29,0xa6,0x14,0x87, 0x82,0xbb,0xae,0x53,
  0x9d,0x51,0x2d,0xfd, 0x24,0xd5,0x98,0x4c, 0x23,0x54,0xf5,0xfd, 0x5f,0x76,0x30,0xa9,
  0xb9,0xac,0xef,0x4c, 0x13,0x08,0x8c,0xe0, 0xca,0x4f,0x03,0x49, 0xd2,0x9c,0xd9,0x6c,
  0x52,0x1a,0x27,0x75, 0x94,0x68,0xce,0xe6, 0x16,0xc9,0xdf,0x54, 0x29,0xf4,0x7b,0x68,
  0xd8,0x98,0x61,0x86, 0x15,0x31,0x53,0x3f, 0x1e,0x0b,0x82,0x1b, 0x60,0x98,0x31,0x48,
  0x44,0x63,0x16,0x61, 0x48,0x64,0xf9,0x77, 0x18,0x3c,0xae,0x6d, 0xfa,0xcb,0x27,0xbd,
  0x29,0x8c,0x45,0x3b, 0xbc,0x74,0x17,0xba, 0xb3,0x6e,0x7d,0x6f, 0x26,0x2f,0x1f,0xa6,
  0x2e,0x45,0x59,0xd4, 0x32,0x96,0x1a,0x88, 0x6b,0xa7,0x1e,0x38, 0x3c,0x29,0xa3,0xf4,
  0xa8,0xe3,0x71,0x68, 0x11,0x3e,0x95,0x01, 0x41,0xcf,0xf0,0xfe, 0x9d,0xe1,0x7f,0xc6,
  // 0x700:
  0xa6,0xfd,0xb8,0xfe, 0x3c,0x15,0x11,0xf7, 0x6e,0xbb,0x88,0x33, 0x83,0x48,0x7e,0xa3,
  0xe0,0xe5,0x8b,0xf5, 0x34,0x82,0x67,0x41, 0xf6,0x5b,0x2f,0x00, 0x61,0x0b,0x09,0x79,
  0x20,0xe1,0x23,0xaf, 0x4f,0xc7,0x97,0xaa, 0x29,0x95,0x0f,0xa7, 0x7e,0x68,0xf0,0x2a,
  0x12,0x4c,0x24,0x92, 0xba,0xd8,0x19,0xbd, 0xc3,0x53,0xa2,0x89, 0x69,0x93,0x0a,0xf1,
  0xd6,0x4f,0xd5,0x78, 0x3b,0x5c,0xcd,0x10, 0xf1,0xa5,0x91,0x26, 0x95,0x4d,0x4f,0x0d,
  0x96,0xf8,0x71,0x61, 0x79,0x34,0x44,0x40, 0x58,0xe9,0x40,0x64, 0xec,0xcf,0xf6,0x2e,
  0x1c,0x19,0x2f,0x47, 0xc3,0x83,0x13,0x09, 0x14,0x8a,0xa9,0xaa, 0x1b,0x9d,0xbe,0x6c,
  0x8a,0x54,0x02,0xc4, 0x9b,0x59,0x0a,0x60, 0x6a,0x91,0x42,0x28, 0x01,0x4c,0x23,0xea,
  0xd0,0xbc,0xd1,0x4c, 0x24,0x14,0x87,0x93, 0xbb,0x8d,0xf4,0xd6, 0x9d,0xf8,0x21,0x5a,
  0x54,0x80,0xa1,0xb8, 0x4c,0x90,0xb6,0x18, 0x33,0xa8,0x52,0x0a, 0x4b,0x7f,0x77,0x7d,
  0x4d,0xe9,0xe9,0x66, 0x9a,0x54,0x26,0x8c, 0x59,0x93,0xd1,0xd1, 0xfb,0x2d,0x99,0xe7,
  0x50,0x3c,0x28,0x1a, 0xa1,0x6a,0xcf,0x67, 0xa6,0xa4,0xde,0xca, 0x61,0xfd,0x0e,0x07,
  0x85,0x56,0x66,0x0d, 0xfd,0x6e,0x7f,0x5f, 0xcc,0x70,0xe4,0x21, 0xb4,0x5d,0x89,0x98,
  0xec,0xea,0x76,0xde, 0x9d,0x8f,0x65,0x34, 0xa7,0x9b,0xe0,0x67, 0x16,0x2d,0xf8,0x28,
  0x2d,0x37,0x08,0xbd, 0x76,0x51,0x35,0xdb, 0x52,0x43,0xcc,0xc3, 0xa7,0xd5,0x40,0x65,
  0x41,0x43,0x2f,0xd4, 0x40,0xc8,0x20,0x80, 0xb0,0x43,0x8d,0xfb, 0xaf,0x57,0x43,0x79,
  // 0x800:
  0xdd,0x5f,0x6e,0x04, 0xe9,0x85,0x78,0xc5, 0x25,0x00,0x90,0x88, 0xee,0x54,0x09,0x11,
  0xdd,0xea,0x2a,0x76, 0x08,0x14,0x94,0x19, 0xcf,0xe2,0xa8,0x2e, 0x13,0x2d,0x0a,0x0a,
  0x81,0x9e,0xa5,0x14, 0xdc,0xb5,0x98,0x8e, 0xb2,0x13,0xef,0x66, 0x51,0xd9,0x8e,0x0c,
  0x36,0x52,0xa4,0x84, 0x63,0xb7,0x6e,0x37, 0x38,0xdb,0x52,0xb1, 0x3a,0x88,0xd5,0x25,
  0x59,0x9c,0x83,0xb9, 0x9d,0x9a,0xca,0x02, 0x9d,0x53,0x9a,0x4f, 0x79,0x2f,0xc9,0x58,
  0x64,0xee,0x61,0x0a, 0x4c,0xfb,0x68,0xbd, 0xbe,0xbc,0xb1,0xe8, 0x49,0x4e,0x9a,0x5f,
  0x96,0xa7,0x8a,0xce, 0xa3,0x90,0xf0,0x38, 0xd9,0x50,0x49,0x7a, 0x84,0xb8,0x80,0x48,
  0x4f,0x57,0xa9,0x3d, 0x9e,0xe8,0xf2,0xf8, 0x7d,0x8f,0x27,0x99, 0xfe,0x00,0x55,0x52,
  0xae,0x56,0xf7,0x7b, 0x3b,0xce,0x56,0x9e, 0x79,0x6a,0xe5,0x23, 0x15,0x7f,0x19,0xbc,
  0x24,0xf2,0x3c,0x91, 0xf8,0x97,0xd6,0x8f, 0x96,0x98,0xd1,0x28, 0x60,0x8b,0x90,0xf1,
  0xa4,0x3e,0x89,0xe4, 0x1d,0x35,0x4b,0xd5, 0x0e,0xe2,0x8a,0x43, 0xdb,0x5f,0xce,0xbd,
  0xa6,0x9d,0x5a,0xfb, 0xa7,0x82,0xd5,0x04, 0xd9,0xe2,0x66,0x5c, 0xc8,0xac,0x86,0xe8,
  0x56,0x20,0x92,0xe9, 0x0e,0x0e,0x2b,0x82, 0x86,0xbb,0x9b,0x8a, 0x09,0x7c,0xd5,0x82,
  0x94,0x4b,0xb6,0x13, 0xaa,0xa7,0x71,0x52, 0x12,0x09,0x07,0x1f, 0xf4,0x69,0xf8,0xe7,
  0x81,0x25,0xf4,0x67, 0x7c,0x4e,0x98,0x2a, 0x8b,0x88,0x4c,0x26, 0xb4,0x3d,0x3d,0x1d,
  0x77,0x58,0xe1,0x38, 0xea,0x68,0x34,0x18, 0xe1,0x72,0xdb,0x7d, 0x38,0x9b,0x2b,0xcd,
  // 0x900:
  0x02,0x32,0x8a,0x1e, 0x44,0x61,0x5a,0x33, 0x24,0x11,0xd6,0xa3, 0xa8,0x7d,0x15,0xc1,
  0x55,0xf5,0x79,0xc6, 0x49,0x96,0xab,0x47, 0xb6,0x3f,0x6f,0xb9, 0x1f,0xee,0x4b,0x7d,
  0xdd,0x32,0x66,0xba, 0xaf,0x77,0x2c,0x54, 0x09,0xfd,0xc0,0xfa, 0xbc,0xb8,0x7c,0x2b,
  0x9d,0xde,0x7c,0x52, 0x86,0xba,0x55,0xcd, 0xa2,0xef,0x3c,0x96, 0x7d,0x1c,0x7a,0x3f,
  0xb6,0x39,0x85,0x23, 0x3a,0xef,0xaf,0x56, 0xad,0xcf,0x19,0xe9, 0x9b,0x1b,0x26,0xb0,
  0x45,0x23,0x66,0xe0, 0x30,0xc2,0xdd,0xb0, 0x48,0xe0,0xa9,0xa4, 0x53,0x1a,0xb0,0x12,
  0xb3,0xd7,0x73,0x4b, 0x0d,0xc2,0x10,0x5a, 0x47,0xf3,0xd8,0x1e, 0x8b,0xf0,0xdb,0x72,
  0x5b,0x84,0x67,0xd9, 0xf0,0x6c,0x57,0x0e, 0xc5,0xe5,0xdf,0x97, 0x6e,0xdc,0x70,0x3d,
  0xcb,0x6c,0x5e,0x87, 0x2c,0xb1,0xec,0xaa, 0x09,0x7c,0x2f,0xcb, 0x82,0xf0,0xdd,0x64,
  0x59,0x3e,0x96,0xeb, 0xdc,0x81,0x27,0x13, 0x85,0x4c,0xc4,0x25, 0x81,0xc0,0x66,0xba,
  0xf7,0x87,0x35,0xbd, 0x96,0x9b,0x59,0xcf, 0x22,0x7e,0xcd,0xbc, 0x56,0x48,0xdb,0x4b,
  0x70,0xc1,0xd7,0xd2, 0x0f,0xb5,0x5a,0xac, 0xfb,0x65,0xb9,0x5e, 0xb9,0xe8,0xec,0x8d,
  0xa5,0xe3,0x6b,0x2d, 0x66,0x60,0xd9,0x64, 0x66,0x70,0x93,0x26, 0x9e,0x3c,0x34,0xdb,
  0x44,0x05,0xdb,0xb3, 0x0a,0x27,0xa2,0xb9, 0x58,0x28,0x72,0xdb, 0x6b,0x58,0xaa,0xe7,
  0xcc,0xb6,0xc8,0x1e, 0xcf,0x6c,0x6c,0x20, 0xab,0x24,0x85,0x43, 0x4a,0xb0,0x20,0x40,
  0xee,0x9e,0xc5,0xdf, 0x8b,0x4f,0xf2,0xd3, 0x6f,0x19,0xac,0x6d, 0xdc,0x57,0x7c,0x4b,
  // 0xa00:
  0x7c,0xd9,0x71,0xb8, 0x36,0xfc,0x2d,0xcb, 0x92,0xad,0x7b,0xa1, 0xb6,0xd8,0xe8,0x5e,
  0x5c,0x39,0x38,0xd3, 0xbb,0x9a,0x48,0xa5, 0xc9,0x3d,0xd9,0x31, 0xe3,0xdd,0x10,0xcd,
  0x9f,0x32,0xee,0xb2, 0xae,0x56,0x96,0x53, 0xcd,0xc0,0x64,0xe5, 0xe3,0x0d,0x58,0x68,
  0x80,0xc5,0x7c,0x89, 0xb0,0x6c,0x19,0xd7, 0x86,0x92,0x43,0x3c, 0x68,0x74,0xe7,0x6b,
  0xf8,0xc9,0x20,0x3a, 0x72,0xb3,0xe0,0x23, 0x82,0x8c,0xcc,0xf6, 0xd2,0x95,0xd3,0xca,
  0x5d,0x47,0xbd,0x2b, 0x36,0xd9,0xb5,0xcb, 0xaf,0x87,0x5a,0x17, 0x6f,0x47,0x44,0x86,
  0x28,0xe6,0xe1,0xba, 0xb4,0xe5,0x57,0xd9, 0xef,0x71,0x14,0x53, 0xc2,0xbb,0x75,0xe5,
  0x17,0xd9,0x27,0x5e, 0xed,0xe2,0x9d,0x09, 0xa3,0x0d,0x91,0x54, 0x36,0x64,0xbb,0x12,
  0xb8,0x71,0xc7,0x7e, 0xec,0xd1,0x4d,0x93, 0x4c,0x5d,0x7c,0x64, 0x53,0x93,0x30,0x58,
  0xec,0x91,0xc9,0xbe, 0x20,0xdb,0xe8,0xc9, 0x87,0x08,0xe2,0x99, 0x55,0xc0,0x64,0xda,
  0xe0,0xec,0xdb,0x31, 0x56,0x69,0xbe,0x5d, 0xbb,0xa5,0x9e,0x60, 0xaf,0x18,0x93,0x90,
  0x31,0x49,0xb9,0x5e, 0x8d,0x09,0xab,0xc2, 0x30,0x19,0x03,0x14, 0x51,0x8b,0x28,0x57,
  0x5f,0xc3,0x8e,0xe7, 0xc4,0x04,0x0d,0x32, 0xbb,0x73,0xa4,0xba, 0xa1,0x65,0x22,0x94,
  0xad,0x31,0x3c,0x4a, 0x1e,0xcb,0xac,0xb7, 0x48,0x73,0x47,0xa3, 0x19,0x7b,0xc6,0x16,
  0xba,0xdc,0xba,0x91, 0x25,0x7d,0x5a,0x16, 0x85,0x26,0xd3,0x39, 0x6c,0x0e,0xfe,0x88,
  0x30,0xa3,0xd0,0xba, 0xb8,0x6e,0xb5,0xbe, 0x9e,0x5f,0x1d,0xee, 0x73,0xcd,0xf5,0xe3,
  // 0xb00:
  0xbb,0x39,0xc5,0xdf, 0x6f,0x1d,0x88,0x2e, 0x5a,0xd1,0xc0,0x52, 0xa6,0xa2,0x03,0xa1,
  0xaf,0x48,0xdf,0x57, 0xda,0xb1,0xdd,0xe6, 0xeb,0x9b,0xed,0x8e, 0xa2,0xa6,0x45,0xab,
  0xd3,0x6d,0xa7,0x08, 0x21,0x46,0xd4,0x9a, 0xc4,0x2c,0x89,0x9e, 0x27,0xb8,0xb1,0xc8,
  0xe5,0x5a,0xac,0x17, 0x53,0x6c,0xed,0x6c, 0xa3,0x4e,0xa3,0x7b, 0xfa,0x77,0x46,0x3d,
  0x3e,0x4e,0xb6,0xcd, 0xf4,0xdb,0x93,0x4c, 0xf3,0x49,0x9c,0x1f, 0x6a,0xe5,0xc6,0xf6,
  0x63,0x13,0xd1,0x10, 0x4d,0x59,0x8c,0x4a, 0xab,0xe6,0xb0,0xc3, 0xaf,0x51,0x93,0x96,
  0x99,0x59,0x73,0x46, 0x46,0x9a,0x78,0x7d, 0x99,0x3b,0x28,0x1e, 0xc4,0xde,0x21,0x37,
  0xb5,0xc9,0xcc,0xf6, 0xbd,0x32,0x38,0x99, 0x1d,0x1c,0x8e,0x3b, 0x13,0xad,0xdc,0x95,
  0x2c,0x40,0x27,0x45, 0xa9,0x76,0xf9,0x6d, 0x93,0x09,0xf3,0xd3, 0x15,0x0e,0x88,0x2e,
  0x30,0xa2,0x80,0xd4, 0xb0,0x56,0xb0,0x01, 0x1e,0x46,0x62,0xfd, 0x9c,0x7b,0x83,0xb2,
  0x4d,0xa2,0x34,0x9c, 0x46,0x61,0xd6,0x96, 0x78,0xe6,0xec,0xfd, 0x09,0x4a,0x20,0x53,
  0x6c,0x5a,0x6b,0x5c, 0x2d,0x24,0xb9,0x8b, 0x5e,0xed,0x6e,0x13, 0xa5,0xc5,0x82,0x1b,
  0x41,0xc7,0xaa,0x7d, 0x39,0x15,0x0d,0xaa, 0x30,0x98,0xcc,0xde, 0x32,0xbd,0xab,0x3e,
  0xd5,0x63,0x17,0x77, 0xdb,0x35,0x7a,0x5a, 0xa7,0xbb,0xb4,0xe5, 0xb0,0x43,0x52,0x9b,
  0xcd,0x2a,0xe7,0xa7, 0x8a,0x4c,0x35,0xc5, 0x68,0xcd,0x81,0x21, 0x0f,0x71,0xd4,0x62,
  0x21,0xe6,0x62,0x99, 0xca,0xcf,0xa9,0x80, 0x9a,0x8b,0xec,0x5e, 0x0e,0x9a,0x06,0x08,
  // 0xc00:
  0x8c,0x42,0xa5,0x8b, 0x25,0x77,0x3d,0xa4, 0xdb,0xc9,0xb9,0x01, 0x35,0xec,0x6f,0x9a,
  0x69,0xd7,0x42,0x0b, 0xcd,0x11,0x18,0x27, 0xbe,0x1c,0xea,0xae, 0xbe,0xa0,0xdf,0x8c,
  0xb1,0xbf,0x47,0xdb, 0x77,0x07,0xb8,0xed, 0xa1,0xa6,0x6a,0xd6, 0xb3,0xa2,0xdb,0x8f,
  0xce,0x25,0x63,0x61, 0xee,0x22,0xc8,0x99, 0x78,0xc3,0xda,0x4a, 0x80,0x97,0x6a,0xa8,
  0x9d,0xc1,0xd7,0x2a, 0x0e,0x32,0xe2,0xfb, 0x4d,0x86,0xed,0x15, 0xbd,0x2a,0x3e,0x33,
  0xe6,0xf2,0x1d,0x64, 0x47,0x6b,0xc8,0xf8, 0x72,0x99,0x95,0xfe, 0x7f,0xc1,0x06,0xc4,
  0xbb,0x66,0xc0,0xce, 0x81,0x9d,0x86,0x52, 0xf2,0xf1,0x24,0xa2, 0x19,0xdd,0x45,0x4e,
  0x86,0xff,0x4b,0xf1, 0x76,0xc9,0x60,0x46, 0x51,0xc7,0xf2,0xf3, 0x10,0x29,0x29,0x1c,
  0x22,0x5e,0x0c,0x36, 0x1d,0xfe,0x1d,0xf7, 0x4b,0x39,0x13,0x43, 0xe6,0x6d,0x5f,0xb1,
  0x37,0xac,0xb6,0x9c, 0xe2,0xf7,0x0c,0x5e, 0xef,0xa1,0x5a,0x7a, 0xb5,0x80,0x90,0x83,
  0x79,0x1b,0x1b,0xcf, 0x57,0x98,0xcc,0x11, 0xbb,0xcb,0x90,0x35, 0xc7,0x92,0x07,0x34,
  0x45,0x8f,0x93,0xae, 0xbe,0x9f,0x40,0x06, 0x6b,0xfb,0x0e,0x33, 0x0c,0x13,0x80,0xec,
  0xbf,0xee,0x7c,0x71, 0xcb,0xf9,0xb2,0xde, 0x3c,0xd6,0x7b,0xcf, 0x92,0x5e,0x7f,0x58,
  0x8f,0x0e,0xbc,0x70, 0x36,0x31,0x80,0x42, 0xfd,0x0f,0x26,0x42, 0x44,0x5b,0xbb,0xb2,
  0x19,0x92,0x63,0xf6, 0x42,0x6d,0x1e,0x45, 0x14,0xe3,0xd3,0x6a, 0x73,0x97,0xd0,0x7d,
  0x8c,0xd8,0xf2,0x9f, 0x43,0x25,0xe3,0xf2, 0x45,0x98,0xd9,0x92, 0x78,0xe0,0xfe,0x11,
  // 0xd00:
  0xff,0x52,0x25,0x28, 0x4f,0xe2,0xc0,0xc9, 0x1a,0x3c,0xc8,0x82, 0x95,0x2a,0x82,0x81,
  0x49,0x9f,0xa4,0x01, 0x27,0x83,0x95,0x72, 0x8c,0x00,0x91,0xf8, 0x00,0x24,0xfe,0xbb,
  0xfd,0x6f,0xdc,0xf2, 0xae,0xc3,0x43,0xd1, 0x9e,0xbc,0x73,0x32, 0x31,0xf0,0x9e,0x9b,
  0xaa,0xcf,0x1c,0xc8, 0xda,0xef,0xa3,0x4a, 0xd1,0xe0,0xe0,0x33, 0xee,0x39,0xfb,0xa7,
  0x77,0xaa,0xf0,0x32, 0x3d,0xbf,0xbd,0x81, 0xcd,0x6a,0xd8,0x34, 0x71,0x4f,0x67,0x3e,
  0xe3,0xfd,0x15,0xc0, 0x55,0xc7,0xec,0x28, 0x16,0x1c,0xad,0xf6, 0x7c,0x35,0xfe,0x8f,
  0xe9,0xe8,0xfd,0x47, 0x1a,0x20,0x4e,0xe9, 0x09,0x61,0xc0,0x0f, 0x7d,0x16,0x93,0x60,
  0xac,0x1f,0x7e,0x28, 0xb2,0x80,0x98,0xf8, 0xec,0xd3,0xe6,0xd4, 0x20,0x2e,0xba,0x77,
  0x96,0xeb,0x93,0x7e, 0x66,0x49,0xa0,0x80, 0xa0,0x25,0x66,0xe5, 0x60,0x05,0xb5,0xef,
  0x7c,0x5a,0x62,0x69, 0xbe,0xa9,0x91,0x9c, 0x34,0xe1,0xb7,0xab, 0x29,0x9b,0x15,0xdd,
  0x54,0x67,0x65,0x36, 0x91,0x9d,0xaa,0xea, 0xd6,0x70,0xd1,0x43, 0x23,0x02,0xdd,0x34,
  0x09,0x91,0xaa,0x78, 0x13,0xa6,0xbe,0x7e, 0x43,0x3f,0x54,0xb0, 0x4d,0xd4,0x54,0x75,
  0x42,0x36,0x8c,0xdb, 0x2a,0xb4,0xfb,0x19, 0xd1,0x04,0xe6,0x38, 0xb4,0x53,0xa5,0xd8,
  0x07,0xb3,0xa1,0xd0, 0x46,0xc2,0x32,0x18, 0xc2,0x5e,0x50,0x41, 0x91,0x14,0x1b,0xeb,
  0x2a,0xc1,0xb2,0x96, 0xa8,0x34,0x19,0xe5, 0x29,0xa4,0x47,0x75, 0x93,0xaa,0x0b,0xf2,
  0x01,0x89,0x90,0x54, 0xaa,0x41,0x41,0xc4, 0x85,0x43,0x49,0xe2, 0x43,0x6e,0xbd,0xa6,
  // 0xe00:
  0x32,0xdb,0x55,0x76, 0xe0,0x15,0x1d,0x63, 0xbf,0x7f,0xa3,0x78, 0x1c,0x95,0x00,0x48,
  0x31,0xb0,0x42,0x93, 0xb8,0x88,0xed,0x47, 0x1e,0x0e,0xc1,0x4d, 0x51,0x78,0x95,0xe1,
  0x76,0xee,0xc3,0x63, 0xe5,0xbb,0xa3,0x38, 0x76,0xe9,0xeb,0x99, 0xbb,0x30,0x65,0x37,
  0x77,0x83,0xce,0x3d, 0xb8,0x93,0xbd,0x3e, 0xed,0xdc,0xdb,0xf1, 0x0f,0x0e,0xaa,0xdd,
  0x3d,0x92,0x82,0x14, 0xeb,0xa5,0x19,0x42, 0x94,0xc5,0xc8,0x0a, 0x60,0x71,0x98,0x7d,
  0xf1,0xd2,0x42,0xc2, 0x9d,0xa2,0x05,0x77, 0x0a,0xc9,0x2e,0x8a, 0x99,0x47,0x99,0x31,
  0xed,0x6b,0xac,0x02, 0x98,0x8a,0xb8,0x38, 0x71,0x3c,0xda,0x96, 0xa6,0xfb,0x37,0xbc,
  0x1c,0xb8,0x45,0xa5, 0x21,0x21,0x9c,0x5e, 0xde,0x22,0x32,0xa1, 0x14,0x53,0x92,0x19,
  0x93,0xac,0x5e,0xc1, 0x04,0xb8,0x38,0x0e, 0x4c,0xca,0x03,0xdd, 0x70,0xc4,0x70,0x75,
  0x1d,0xf8,0x40,0xd7, 0x15,0x32,0x36,0x12, 0x1a,0x15,0x10,0x2c, 0x38,0xc7,0x53,0xb3,
  0x0f,0xd2,0xf0,0xdc, 0xe8,0xb3,0xc0,0xb8, 0x7b,0x31,0x68,0x7e, 0xa4,0x4e,0xe6,0x9c,
  0x04,0xfb,0x0c,0x0d, 0xbf,0x86,0x0e,0xc2, 0x95,0xb0,0xef,0xe5, 0x53,0x55,0x99,0x33,
  0xa0,0x66,0xc3,0x71, 0x60,0x85,0x99,0xbb, 0xb6,0xae,0x8a,0xd9, 0xd6,0xbc,0xac,0x56,
  0x14,0x61,0x10,0x21, 0xf7,0xcd,0x97,0x58, 0x61,0x21,0x96,0xbc, 0x64,0x84,0xcb,0x2a,
  0xc8,0xa1,0x6e,0xc9, 0x76,0x95,0x10,0x55, 0x81,0x1a,0x91,0x31, 0x9b,0xe7,0xc7,0x9b,
  0x19,0xc7,0x85,0xf0, 0xc6,0xa8,0xa0,0xb4, 0x55,0x75,0x4e,0xf4, 0xc6,0x84,0x26,0x1e,
  // 0xf00:
  0x5b,0x2e,0x65,0x6b, 0x4d,0x14,0xcc,0x12, 0xda,0xbb,0x8e,0xae, 0xa4,0xc5,0xb4,0x24,
  0x7a,0x7b,0x97,0x72, 0x61,0xd0,0xd2,0x81, 0xd8,0x75,0x79,0x3f, 0xa1,0x47,0xd1,0x96,
  0x72,0x1a,0x93,0x58, 0x99,0x1e,0x0b,0xc5, 0x54,0xb7,0xb0,0x80, 0x80,0x97,0x0d,0x6a,
  0x60,0x94,0x2f,0x29, 0x24,0x31,0x3b,0x0f, 0x58,0x4e,0x09,0x3a, 0x94,0x53,0x3b,0x92,
  0xee,0xc3,0xfb,0x83, 0x30,0x86,0x02,0x70, 0x21,0x68,0x6a,0x60, 0x1a,0x7c,0x48,0x30,
  0x26,0x9a,0x49,0x21, 0x4a,0x60,0x92,0x24, 0x06,0x24,0x88,0x60, 0x84,0x98,0x92,0x4a,
  0x01,0xe3,0x32,0x98, 0x83,0x70,0x28,0x87, 0x03,0x19,0x76,0xf8, 0x4e,0x04,0x87,0x57,
  0x0e,0x24,0x37,0x51, 0xc7,0x9b,0xa3,0x04, 0x06,0x3a,0x8c,0xab, 0x59,0x62,0xa5,0x04,
  0xe1,0x38,0x8c,0x10, 0xdb,0xdb,0x5c,0x0c, 0x1a,0xca,0x9e,0x1d, 0xa3,0x88,0x03,0xa9,
  0x37,0x5c,0xbb,0x1f, 0x46,0x53,0x38,0xda, 0xef,0x62,0x89,0x66, 0xd9,0x2f,0x77,0xb4,
  0x8c,0xa6,0xf5,0xa0, 0xf8,0x36,0xfe,0xa3, 0x32,0x8e,0x0e,0xcc, 0xfe,0xbe,0x27,0xd0,
  0x3d,0x4f,0xeb,0xb3, 0x67,0x87,0x18,0xad, 0x5c,0xf9,0xed,0x5b, 0xc3,0x35,0x23,0xa2,
  0x94,0x1e,0x42,0x9d, 0x83,0xce,0x83,0x9a, 0x34,0xa4,0xd1,0x8c, 0x13,0xa8,0x46,0x9c,
  0x8e,0x78,0x1a,0x77, 0xb3,0x82,0xa9,0x90, 0xdb,0xc8,0x60,0x88, 0x7b,0x79,0xd5,0x07,
  0x97,0xb6,0x25,0x6e, 0xe1,0x80,0x6a,0x16, 0xa6,0xa5,0x42,0x25, 0x90,0x90,0x98,0x51,
  0x85,0x82,0xa0,0x4a, 0x94,0xe5,0x83,0x41, 0xdc,0xc4,0xd8,0xe1, 0x98,0xd1,0x3e,0xc2,
  // 0x1000:
  0xc2,0x69,0x08,0x87, 0x75,0x8a,0x4a,0x1b, 0xe5,0x29,0x0f,0x35, 0xa3,0x12,0x01,0x04,
  0x88,0xc4,0x25,0x30, 0x96,0xc0,0xb1,0xbb, 0xf2,0xd8,0x5a,0x01, 0xae,0x3c,0x26,0x02,
  0xbc,0xb5,0xde,0xe6, 0x03,0xd9,0x52,0x67, 0x6d,0xe6,0xe0,0x39, 0x36,0x6c,0xae,0xb4,
  0x90,0xd1,0x0f,0xb3, 0x4e,0xdb,0xa0,0xa9, 0x5c,0x8e,0x38,0xa6, 0x50,0xfc,0xb6,0x63,
  0x21,0xca,0xe0,0x64, 0xed,0x7c,0xc9,0xe7, 0xf9,0x7b,0x28,0xf3, 0xe0,0xe8,0x4d,0xac,
  0x14,0xba,0xd2,0xc8, 0x75,0x44,0x40,0x54, 0x32,0x1c,0x91,0xb4, 0x3a,0x64,0xa0,0xf3,
  0xd0,0xe4,0x06,0xa1, 0xc2,0x38,0xdc,0xba, 0xc5,0x59,0x12,0x1e, 0xd4,0x2b,0xa9,0x1d,
  0x2a,0x01,0xe1,0x70, 0x3c,0xdf,0x7b,0xbb, 0xe9,0xfa,0x7d,0x15, 0xec,0xb3,0x97,0xaa,
  0x70,0xbe,0xef,0x51, 0xfd,0x16,0x94,0x94, 0x4d,0x75,0xf9,0x2a, 0x09,0x87,0xd7,0x11,
  0xc3,0xde,0x7f,0x9b, 0x8c,0x9e,0x85,0xab, 0x7c,0xa7,0xd0,0xa0, 0x85,0xe5,0x9e,0x5b,
  0x3d,0x0f,0x8e,0x2e, 0xd0,0x95,0xd5,0x94, 0xf4,0xca,0x74,0x80, 0x47,0xec,0x23,0x19,
  0x8c,0x53,0xf1,0x1b, 0x91,0xda,0x1a,0x3a, 0xc5,0x2c,0x3c,0x9a, 0x91,0x02,0x2b,0x51,
  0x0c,0x21,0x0c,0xda, 0xc6,0x79,0xb9,0xd7, 0x40,0x00,0x67,0x0e, 0x30,0xde,0x0d,0x6d,
  0x0e,0xf3,0xe7,0x91, 0x8d,0x23,0xdc,0x5d, 0x6b,0x87,0x0c,0x42, 0x5c,0x53,0x5f,0x6e,
  0xfa,0x09,0xf0,0xb2, 0x7e,0x48,0x70,0x7d, 0xdf,0x3e,0x27,0xd5, 0x59,0x68,0xfe,0x5b,
  0x13,0xd9,0x55,0x1c, 0x20,0x71,0x42,0xd9, 0xed,0xf1,0x5c,0x34, 0x71,0x35,0x33,0x3a,
  // 0x1100:
  0x2e,0x79,0x38,0x5f, 0x0c,0x69,0x1c,0xf4, 0xbb,0x3a,0xf9,0xd7, 0x5c,0x3b,0x59,0x24,
  0xad,0x2f,0x3f,0x69, 0x3d,0x28,0x40,0x90, 0xb2,0x1c,0x2c,0x42, 0x48,0x68,0x22,0x96,
  0x83,0x52,0x75,0xf7, 0xb6,0xe7,0xd1,0x9b, 0x45,0x86,0x26,0x45, 0x3d,0x79,0x81,0x6a,
  0xc4,0x22,0x70,0xd6, 0x2e,0x95,0x95,0x41, 0x4c,0x16,0x7c,0x14, 0x10,0xab,0x19,0x0c,
  0x26,0xdb,0xaf,0x58, 0x84,0xd6,0xf0,0xed, 0x70,0x1e,0xfd,0x4f, 0x82,0x3d,0x71,0xee,
  0xf0,0xb5,0x65,0x31, 0x4b,0x9e,0xd0,0x40, 0xdd,0x13,0x52,0xb7, 0x1e,0x50,0xc4,0xa1,
  0x71,0xbe,0x38,0xcb, 0xee,0x8e,0xad,0x1c, 0x30,0xc1,0xe1,0xec, 0x70,0xed,0x30,0xd5,
  0x0c,0x7b,0x1a,0x46, 0xb3,0xd6,0x9f,0x4c, 0x43,0x52,0x04,0x2e, 0x9b,0xf7,0xc1,0x7a,
  0xf7,0x5b,0xdc,0x96, 0x27,0x31,0x50,0xae, 0x11,0xf6,0xf0,0x64, 0x96,0x0a,0xa3,0xb6,
  0xaa,0x9d,0x1c,0x58, 0x47,0x6e,0x26,0x6a, 0x15,0x2b,0x05,0x06, 0x82,0x49,0x42,0x10,
  0x10,0x04,0x84,0x30, 0x4d,0x30,0x9c,0xd6, 0x10,0x34,0xc4,0x94, 0x44,0x50,0x25,0x45,
  0x44,0xd2,0x45,0x13, 0x52,0x11,0x44,0x81, 0xd3,0x06,0x24,0x11, 0x34,0x4c,0xcc,0xde,
  0x17,0x6f,0x4b,0x6f, 0xe5,0xd6,0xea,0x90, 0xa6,0x93,0x49,0x26, 0x12,0x07,0x54,0x76,
  0xe7,0xb1,0x78,0x05, 0xed,0x48,0xcb,0x87, 0x20,0x62,0x4a,0x48, 0xc0,0x51,0x2c,0xd2,
  0x73,0xc6,0x70,0xaa, 0x00,0xde,0x98,0xac, 0x5e,0xe3,0x2a,0x05, 0x14,0x60,0xa4,0x14,
  0x54,0x48,0x89,0x10, 0x41,0x62,0x4d,0x19, 0x0e,0x0c,0x03,0x45, 0x6c,0x49,0xda,0x32,
  // 0x1200:
  0x63,0x9e,0xcb,0xef, 0x42,0x93,0x0c,0x39, 0x26,0xd1,0x60,0xe6, 0x3d,0x21,0x8f,0x7f,
  0x1d,0xd0,0x69,0xd5, 0x8a,0x3a,0xaa,0x51, 0x85,0xa2,0x67,0x41, 0xc9,0x31,0x06,0xee,
  0xd0,0x1b,0x28,0xcb, 0xc9,0x76,0x23,0x9e, 0xcc,0xde,0x5b,0xf2, 0x26,0xd6,0x16,0xcf,
  0x2a,0x6d,0x4d,0x53, 0x44,0xb4,0xe2,0xf1, 0x18,0x5c,0xd2,0x80, 0x78,0x15,0x21,0x77,
  0x53,0xb1,0x38,0x46, 0x1c,0x19,0x41,0xe2, 0xac,0xb1,0x53,0xd3, 0xe3,0xa8,0x58,0xe8,
  0xc0,0xd5,0x37,0x62, 0x6b,0xaa,0x32,0xce, 0xdd,0x76,0xa8,0x21, 0x75,0x58,0x73,0x6b,
  0x8e,0x18,0xec,0xe7, 0x0e,0x8b,0x28,0x43, 0x9d,0xb0,0x0a,0xa1, 0x5a,0x3f,0x2b,0x19,
  0xed,0x84,0x11,0xac, 0xed,0x16,0xcd,0xc9, 0xc4,0x74,0xf2,0xba, 0x49,0x22,0x9b,0x6a,
  0x84,0x34,0x3b,0x26, 0x57,0x50,0xd0,0xc1, 0x9c,0x17,0x46,0x2c, 0xc7,0x74,0x0c,0xed,
  0x03,0xf8,0xb7,0xb3, 0xcc,0xb3,0x3f,0x8d, 0x45,0x4b,0x0f,0xf6, 0x9f,0x66,0xab,0xc3,
  0xdd,0xec,0x44,0xc6, 0xc5,0xb4,0x73,0x44, 0x21,0x6c,0xe6,0x2c, 0xe1,0xd3,0xcf,0xdc,
  0x55,0xec,0xb7,0xbb, 0x3c,0x02,0x97,0x7b, 0x20,0x90,0x41,0x8c, 0xb6,0x9c,0x32,0xa7,
  0xae,0xa6,0x0c,0x87, 0x1b,0xaa,0x8c,0x16, 0x5d,0x8c,0xcc,0x5b, 0x6d,0xae,0x02,0x34,
  0x29,0x8e,0xfc,0x63, 0xd2,0x57,0x52,0x38, 0xb4,0x60,0x5d,0x5e, 0x82,0xc2,0xa8,0xd6,
  0x51,0x84,0xb6,0xcc, 0x7a,0x11,0x24,0x09, 0xdb,0xed,0xfd,0xbd, 0xf1,0x95,0x49,0x2d,
  0xb6,0xf7,0xbc,0xb2, 0x15,0x7d,0x3a,0xa0, 0xe8,0xf4,0xcc,0xc7, 0x09,0xa5,0xb6,0xc3,
  // 0x1300:
  0x89,0x67,0xbe,0x6e, 0x0b,0x9f,0xbd,0x05, 0xbd,0xfc,0xb0,0xdf, 0xe0,0x9a,0x50,0x37,
  0x6d,0xc8,0x66,0x6e, 0xc2,0x7b,0xdf,0x1c, 0x19,0xa9,0xd0,0x53, 0xa3,0x98,0x38,0xf2,
  0x3c,0xa6,0x42,0x62, 0xaf,0x00,0xf1,0x54, 0x9d,0xc4,0x76,0xb7, 0x9c,0x78,0x41,0x83,
  0x99,0x17,0x92,0xe9, 0xb6,0x47,0xce,0xe6, 0xe6,0xad,0x9d,0x07, 0x69,0x12,0x3d,0xa7,
  0x2d,0xd2,0x62,0xec, 0x99,0x0f,0x3b,0xb6, 0x0d,0x55,0x07,0xaf, 0xcb,0xd9,0x38,0xf9,
  0xcf,0x63,0x83,0xf8, 0x00,0x5b,0x69,0x43, 0xa4,0xed,0xcc,0x4e, 0xee,0xe2,0x1e,0x7e,
  0xfd,0x4b,0x5d,0x19, 0x5c,0x49,0x86,0x8c, 0x33,0x7c,0x16,0xa6, 0xca,0xd4,0x73,0x91,
  0x9f,0x5f,0x8a,0x67, 0xa3,0x9f,0x74,0x6a, 0xe8,0x8c,0xb0,0x20, 0x51,0xb4,0x51,0x0c,
  0x2d,0x4a,0xec,0x92, 0x24,0x47,0x10,0x15, 0xac,0x71,0x71,0x9d, 0x40,0x9f,0xac,0xf5,
  0x26,0x23,0x14,0x61, 0xde,0x5f,0xef,0xbd, 0xc8,0x0b,0x68,0x8a, 0x62,0xdb,0xb5,0x0b,
  0x4f,0x1b,0xb9,0x96, 0xd4,0xb7,0x51,0xc9, 0x00,0xca,0xfd,0x1c, 0xa9,0xc2,0x6f,0xc1,
  0x43,0x92,0x87,0x8e, 0x9f,0xef,0x28,0xc3, 0x88,0x0a,0x62,0x56, 0x9c,0xb0,0x85,0x63,
  0x31,0x0a,0xc2,0xa4, 0x4e,0x1c,0x5e,0xcc, 0x9e,0x1f,0xad,0x3a, 0x9d,0x1d,0x51,0x3d,
  0x05,0x63,0xd5,0x52, 0x18,0xe8,0x5a,0x3e, 0x73,0x57,0x39,0x97, 0x4b,0x5e,0xce,0xdb,
  0xdf,0x0c,0xe8,0x76, 0xb9,0x6b,0x32,0xcf, 0x3e,0x0a,0xe7,0xe7, 0xfe,0x30,0xd4,0xfd,
  0x42,0x81,0x48,0x2c, 0xfd,0x72,0x3c,0x48, 0x7a,0x81,0xeb,0xd2, 0xc5,0xa0,0x31,0x53,
  // 0x1400:
  0x13,0x07,0x09,0xae, 0x9e,0x79,0xd6,0x67, 0xb6,0x32,0xb2,0x81, 0x77,0x51,0x51,0xd0,
  0x37,0x57,0xe0,0xa4, 0x5b,0x24,0xea,0x43, 0xba,0x9f,0xeb,0x45, 0x92,0x7a,0x26,0x0a,
  0x79,0x90,0xb3,0x0b, 0x2a,0xf5,0x9f,0xa2, 0xb5,0xf6,0x2f,0xa3, 0x5f,0x5e,0xbf,0x36,
  0x3c,0xea,0x73,0x7a, 0xec,0xe6,0x3d,0xb9, 0x96,0x22,0x7c,0x6e, 0x71,0x33,0xce,0xa0,
  0xdd,0x07,0xde,0x5e, 0x39,0x9a,0x00,0x9b, 0x34,0x07,0x8a,0xa7, 0x0e,0xac,0x4f,0x87,
  0xb5,0x67,0xa8,0xce, 0xc9,0x8b,0x9d,0x54, 0xd6,0xa3,0x53,0x3d, 0x6a,0x67,0x08,0x18,
  0xb5,0x93,0x00,0xb7, 0x12,0xf5,0x0d,0xeb, 0x77,0x4c,0xf8,0x0f, 0x16,0xde,0x58,0xc4,
  0x60,0x99,0x26,0x2a, 0xd3,0xb5,0x2b,0x69, 0x47,0x91,0x38,0x50, 0x22,0x2c,0x6a,0x58,
  0x28,0x65,0x52,0xa2, 0x12,0x81,0x23,0xfa, 0xc2,0x28,0x3e,0x31, 0xdd,0xeb,0x3d,0x04,
  0xa4,0x6b,0xe3,0xac, 0xfa,0xf5,0x94,0xce, 0x19,0x27,0xc2,0x79, 0x05,0x07,0x86,0x5f,
  0x4e,0x67,0x17,0x1e, 0x1e,0x60,0x88,0x00, 0x01,0x44,0x6f,0x45, 0x17,0x02,0x1f,0x74,
  0xed,0xdc,0x35,0x6b, 0x5d,0xc7,0xbe,0x53, 0x7f,0xa5,0x3a,0x3f, 0xa1,0xe1,0x7b,0x28,
  0x89,0xe1,0x64,0xd0, 0xae,0x2e,0x87,0x51, 0x64,0xcb,0x5d,0x37, 0x6c,0xf0,0xc2,0x24,
  0x9b,0x99,0x98,0x9d, 0x70,0x3e,0x3a,0xab, 0xa0,0xae,0xc4,0x82, 0x50,0x05,0x02,0x50,
  0x0c,0x49,0x40,0xd0, 0x34,0x94,0x85,0x05, 0x01,0x35,0x01,0x40, 0x14,0xa5,0x0d,0x0d,
  0x23,0x9b,0x36,0x99, 0xd7,0xe7,0xde,0x65, 0x32,0x61,0x0c,0x6a, 0xe5,0x6b,0x1c,0xa6,
  // 0x1500:
  0xde,0x65,0x0c,0x81, 0xf8,0x3d,0x52,0xe7, 0xf0,0x65,0xeb,0x77, 0x36,0x89,0x13,0xa7,
  0x9e,0x39,0x5e,0x3d, 0x7c,0x59,0xe7,0xdf, 0xbf,0x65,0x48,0x28, 0x47,0x1a,0xde,0xbe,
  0x37,0x85,0x42,0x09, 0x4b,0x87,0x60,0xa2, 0xca,0xa1,0x53,0x8e, 0x65,0x0f,0x70,0x93,
  0xa4,0x01,0x19,0x3c, 0xc7,0xb6,0x54,0xa0, 0xaf,0xbd,0x24,0x2e, 0xbb,0x61,0xb8,0xfa,
  0xf9,0xf5,0x46,0xf5, 0x83,0xc5,0x88,0x8d, 0x8b,0xa1,0x30,0x89, 0xe6,0x39,0x35,0x11,
  0x0c,0xc9,0x0e,0xfe, 0xec,0x33,0x55,0xe5, 0x9a,0xa5,0xec,0x64, 0xee,0x22,0x45,0x8a,
  0x7e,0x16,0xcb,0xb1, 0x6f,0x17,0x1b,0xfe, 0x15,0x9f,0x9f,0x06, 0xd3,0xee,0x09,0x2e,
  0x9a,0xdb,0xfb,0xf1, 0x0a,0xb8,0x6a,0xd4, 0xbd,0xe7,0xbd,0x88, 0x05,0x12,0xb5,0x22,
  0x40,0xc1,0x27,0x00, 0x88,0xd7,0x38,0x6e, 0xce,0x12,0x65,0x4e, 0x8c,0x20,0x8e,0xde,
  0x17,0x18,0xc9,0x6c, 0xc9,0x51,0x42,0x9b, 0xa6,0x10,0x29,0x35, 0x32,0xcd,0xaa,0xc5,
  0x91,0xb8,0xb2,0x0e, 0xe1,0x2c,0x16,0xfb, 0xc5,0xf5,0x3c,0x5d, 0x10,0x5d,0x47,0x56,
  0xd4,0xd3,0x76,0xe8, 0x0d,0xb0,0x38,0xbb, 0x36,0x15,0x0e,0x35, 0xf1,0x72,0xc4,0xb1,
  0xa6,0x3b,0xea,0x50, 0x9f,0x83,0xef,0xeb, 0xc0,0xc3,0xa7,0x9d, 0x64,0xad,0xa4,0x68,
  0xec,0xb2,0x99,0x38, 0x7b,0x4f,0x21,0x8e, 0x5e,0x09,0xbd,0x34, 0x6b,0xdf,0x50,0x20,
  0xd5,0x42,0x99,0xb0, 0x30,0x6a,0x94,0x50, 0x59,0x2b,0xd1,0x52, 0xd6,0x0d,0x9e,0x9a,
  0x3a,0x93,0x09,0x1e, 0xbd,0x3b,0x35,0x0d, 0xb5,0x95,0xe3,0x4a, 0xad,0x4d,0x82,0xa9,
  // 0x1600:
  0x21,0x90,0x2a,0x4e, 0xe2,0x05,0xea,0x6a, 0x6f,0x1e,0xb8,0x1e, 0x08,0xb0,0x78,0x9f,
  0x47,0x14,0xfa,0x69, 0xf6,0xe4,0x09,0xdf, 0xf4,0xaf,0x33,0x74, 0x84,0xbd,0x1c,0x0d,
  0xc0,0xca,0x10,0x99, 0x20,0x71,0x3b,0x85, 0x54,0x31,0x04,0xcf, 0x09,0x26,0x38,0xb7,
  0x28,0x1b,0x0b,0xc8, 0xc6,0x0b,0x72,0xd6, 0xdc,0xc8,0xcd,0xbb, 0x64,0x8e,0xab,0x50,
  0x5a,0xe8,0xec,0xf3, 0x17,0xd2,0xa5,0xb5, 0xb9,0xc9,0x99,0x4e, 0xc3,0xcd,0xd2,0xf1,
  0x01,0x15,0x05,0x17, 0x36,0x93,0x80,0x54, 0x3a,0xd9,0x54,0x05, 0x33,0xc8,0xaa,0x6e,
  0x63,0x64,0x13,0x55, 0x0e,0xa8,0x3a,0xe3, 0x7f,0xb4,0xc9,0xb9, 0xd3,0xb6,0x4d,0x91,
  0xd9,0x05,0x0d,0x73, 0xd4,0xdb,0x18,0x5d, 0x99,0x12,0x2a,0x18, 0xd0,0xea,0x86,0xa9,
  0xaa,0x06,0xc7,0x45, 0x35,0x55,0xb1,0xbb, 0xb7,0xc2,0x95,0xbe, 0x5d,0x64,0x38,0xf0,
  0xeb,0xf8,0x3a,0xa0, 0x33,0xc0,0xec,0xb7, 0x6b,0xbb,0x9d,0x18, 0x23,0xe2,0x00,0xf6,
  0x35,0x74,0xbb,0x3d, 0x40,0xa8,0x51,0xda, 0x6e,0xc7,0x2e,0x92, 0x40,0xef,0x1d,0xd4,
  0x5c,0x96,0xf4,0xb0, 0xba,0x57,0x89,0x04, 0x6f,0x24,0xa1,0x04, 0xc4,0xa0,0x61,0xea,
  0xf5,0x30,0x0f,0x15, 0x09,0x7a,0x7b,0x57, 0xbc,0x5a,0x46,0x49, 0xdc,0x15,0x42,0x8f,
  0x21,0xf1,0x55,0x97, 0x99,0x5c,0x61,0x23, 0x72,0xac,0xdd,0x4b, 0x23,0x85,0xe0,0x50,
  0x1c,0x24,0x20,0x91, 0xd8,0x74,0x8e,0x4a, 0x4b,0xe0,0x3a,0xfd, 0xc0,0xb5,0x2b,0x75,
  0x79,0x88,0xa6,0xbe, 0xd8,0x68,0x82,0x33, 0x27,0x23,0x58,0xf2, 0xd0,0xd4,0x89,0xb5,
  // 0x1700:
  0x49,0x17,0x53,0x32, 0xdc,0xd4,0xcc,0xbf, 0x4d,0xe4,0x84,0x70, 0xf1,0xec,0xc9,0x35,
  0xba,0x13,0x37,0x6d, 0x03,0xa8,0x12,0xaa, 0x90,0x40,0x29,0xae, 0x36,0x55,0x85,0x72,
  0x44,0x91,0x14,0x25, 0x3a,0xbb,0xd8,0x04, 0xbc,0x7c,0xce,0x77, 0x33,0xd3,0xb1,0xaf,
  0x51,0xad,0x43,0x3c, 0xb2,0x2c,0x48,0x91, 0x61,0xa1,0xa5,0x4f, 0x35,0x27,0x51,0x4c,
  0x8b,0x55,0xb4,0x17, 0xc3,0xef,0x7b,0xa6, 0xe3,0xa5,0x77,0xfa, 0x7a,0x43,0x79,0xbb,
  0xc7,0x3b,0xdd,0xc5, 0x9f,0xae,0x75,0x4c, 0x34,0xbb,0x38,0x9b, 0x5e,0xe7,0x1d,0x5f,
  0x43,0x58,0x57,0x5b, 0x6c,0xd2,0x67,0x5e, 0x7b,0xad,0xc1,0x2c, 0x18,0x92,0x9b,0x85,
  0xb5,0xdf,0x69,0xbd, 0x2d,0xb4,0xca,0x0e, 0xee,0x1d,0x48,0xa7, 0x22,0xa9,0x03,0x19,
  0xf0,0x11,0x0c,0xf7, 0x35,0xa4,0x38,0xf3, 0x63,0x11,0xb2,0x22, 0x51,0xd9,0xf4,0xb6,
  0x3a,0x80,0x45,0x56, 0x28,0x0c,0x47,0xaa, 0x03,0xcb,0xee,0xaa, 0x0a,0x88,0x24,0x92,
  0x62,0x89,0xa2,0x88, 0x26,0xa6,0x8a,0x60, 0x89,0x18,0x22,0x8b, 0x15,0x8c,0x62,0xac,
  0x71,0xb9,0xc8,0x8f, 0xa3,0xb2,0xb1,0x8f, 0x6b,0x4f,0x3e,0x2f, 0xe7,0x70,0xeb,0x11,
  0x63,0xbc,0x45,0x89, 0x42,0x2f,0x51,0xed, 0x9c,0x96,0xbf,0x79, 0x40,0xe0,0x40,0x3e,
  0x35,0x0a,0x69,0x5e, 0x42,0x90,0x20,0x9e, 0x77,0x1d,0xc4,0x19, 0x0f,0x5f,0x73,0x6f,
  0x77,0xb1,0x87,0xa8, 0xb1,0xa8,0xa8,0xa0, 0xf1,0x70,0xc8,0xbb, 0x59,0x8f,0x77,0x3a,
  0x52,0xe0,0x9b,0xab, 0x03,0x57,0x41,0xef, 0x0e,0xf4,0xbe,0x85, 0xd9,0x73,0x91,0xc7,
  // 0x1800:
  0xfe,0x05,0x20,0xa7, 0x08,0x01,0xe2,0x3f, 0xf0,0xeb,0xa5,0xba, 0xef,0x33,0x21,0x98,
  0x38,0x70,0xea,0x8f, 0x0f,0xb1,0x9c,0x48, 0x88,0xe3,0x19,0x37, 0xb6,0xe3,0x9d,0x43,
  0x8c,0xda,0xac,0xcb, 0xf0,0xbf,0x6d,0xab, 0x53,0xaf,0xb3,0x65, 0xcd,0xe8,0x6a,0x58,
  0x54,0xee,0x4a,0x43, 0x72,0x1e,0x64,0xb4, 0x08,0x39,0x17,0xcc, 0x58,0x9c,0xde,0x9b,
  0x49,0x2b,0xd4,0xa1, 0x2c,0x2b,0xda,0x51, 0x9b,0x45,0x20,0x2c, 0x45,0xc3,0xe4,0x1d,
  0x5d,0x97,0x32,0x6a, 0x51,0xdf,0x3d,0xe1, 0xf0,0x60,0xa2,0x94, 0x8a,0x20,0x25,0xff,
  0xd8,0x61,0x99,0x46, 0x49,0x41,0x45,0x11, 0x53,0x54,0xd4,0x15, 0x98,0xb8,0x45,0x05,
  0x25,0x31,0x1b,0xd3, 0x6e,0x36,0x76,0x6b, 0x7d,0x87,0xda,0xa5, 0xa8,0x8b,0xdd,0xeb,
  0xc7,0x71,0x83,0xc4, 0x86,0x7c,0x2b,0xef, 0x5d,0x35,0xc8,0xb2, 0x3e,0xda,0x29,0xdb,
  0x53,0x55,0xc2,0xea, 0x22,0x84,0x95,0x26, 0x9b,0x37,0x26,0x74, 0x6c,0x5e,0xb8,0x12,
  0xb8,0xd2,0x9b,0x54, 0xd6,0x28,0xb2,0x82, 0x60,0x94,0x9e,0xba, 0x95,0x18,0x6b,0x59,
  0xef,0xbb,0xc0,0xfb, 0x07,0x1f,0x0d,0x61, 0xd8,0x81,0xaa,0x1f, 0x3a,0x7b,0x7b,0x3c,
  0xe6,0x38,0xbb,0x23, 0xe6,0xed,0x2a,0x74, 0xa1,0xc9,0x9b,0x9c, 0xfd,0xad,0x03,0x86,
  0xb1,0xd2,0x54,0x2f, 0xd3,0xbf,0x8e,0x33, 0x72,0x7e,0x4e,0xdb, 0x37,0xe0,0xa7,0x89,
  0xd7,0xa6,0xa6,0x3f, 0xf3,0xe4,0xf3,0xfc, 0x38,0x9c,0x35,0xa0, 0xf2,0xcb,0x35,0x2e,
  0x4d,0xc5,0xd4,0xa0, 0x7b,0x1a,0x4c,0x63, 0x85,0x98,0x13,0xbf, 0xbe,0xa5,0x8d,0xc4,
  // 0x1900:
  0x56,0x2d,0x12,0x9b, 0x0a,0x8e,0xd7,0x3c, 0xf1,0xcd,0x36,0x16, 0x55,0x1c,0x1d,0xd7,
  0xc2,0x9d,0x01,0x22, 0x14,0xa1,0xa4,0xeb, 0xbf,0x93,0xe8,0x6f, 0xc8,0x8b,0x17,0x34,
  0xee,0x4c,0xf0,0xd0, 0x29,0x52,0xaa,0xea, 0x91,0x62,0x53,0x4c, 0xde,0x6e,0xa0,0xb0,
  0x73,0x36,0xee,0x2c, 0xc0,0x2c,0x39,0x35, 0x6e,0x69,0xd4,0x86, 0x15,0x86,0xd9,0xd0,
  0xb2,0x5b,0xa2,0xb9, 0x2c,0x91,0x33,0x36, 0x87,0x1f,0x12,0xae, 0xe0,0x67,0x7a,0xcb,
  0x6c,0x56,0x45,0xf1, 0x72,0x8b,0xb4,0xc1, 0xda,0x07,0x9d,0xa2, 0x58,0xda,0x70,0xad,
  0x37,0xb4,0x0b,0xc5, 0x3e,0xa3,0x2e,0x15, 0x7b,0xc4,0x09,0x8f, 0x20,0xbe,0x0f,0x81,
  0xf4,0x3e,0x2f,0x99, 0x03,0x2d,0x9e,0x45, 0xc5,0x44,0x80,0xd7, 0x0a,0x4b,0xab,0x7f,
  0xba,0xf7,0x96,0x12, 0x05,0xc9,0xab,0x52, 0xc8,0x84,0xa9,0xb1, 0xbd,0x85,0xd7,0x0d,
  0xe0,0x1a,0xe4,0x51, 0x4a,0x4d,0x34,0x31, 0x4d,0x0f,0x42,0xc4, 0x93,0x54,0x9d,0xc8,
  0x83,0xc2,0x82,0x60, 0x28,0xaa,0x68,0x6a, 0x49,0x0a,0x49,0x88, 0x92,0x86,0x20,0x28,
  0xa4,0x92,0xa5,0x22, 0x51,0x88,0x28,0x29, 0x28,0x29,0x4a,0x18, 0xa2,0x06,0x22,0x9a,
  0xaa,0x18,0x20,0x08, 0x96,0x20,0x09,0xa2, 0x52,0x9a,0x58,0x96, 0x94,0x62,0x40,0x26,
  0x1a,0x40,0x4d,0xbd, 0x38,0x50,0x96,0x2d, 0x61,0xab,0x28,0x48, 0xce,0x0a,0xeb,0x23,
  0xa4,0x8e,0xe4,0xba, 0x83,0x24,0x47,0x49, 0xbe,0x55,0x3d,0xe5, 0xf6,0x62,0x45,0x00,
  0x82,0x09,0x04,0x92, 0x0c,0xef,0xf0,0x9a, 0x2a,0x8a,0x51,0x57, 0x16,0x63,0x4f,0x8e,
  // 0x1a00:
  0x61,0xd8,0x72,0x67, 0x8d,0xcd,0x71,0x5e, 0xa5,0x51,0xbc,0xf3, 0xf4,0x66,0x5e,0x5d,
  0x17,0xf1,0xe3,0x37, 0x25,0x47,0xa5,0x94, 0xc1,0xe9,0xbb,0x2f, 0x95,0x02,0xd7,0x1d,
  0xf5,0x60,0xaa,0xbb, 0x39,0xb3,0x6c,0xe1, 0xe9,0xcc,0xab,0x15, 0x40,0xb7,0x18,0x75,
  0xa2,0xaa,0x88,0x34, 0xf4,0x30,0xbd,0x27, 0x99,0xd6,0xbc,0xc3, 0x35,0x87,0xcc,0x61,
  0x36,0xdc,0xf5,0xc6, 0x59,0xb6,0xa9,0x78, 0x3d,0x8c,0x2f,0xe0, 0x7d,0x9b,0x3c,0x59,
  0x66,0x9b,0x84,0xec, 0x56,0x07,0x07,0xb7, 0x7d,0x0a,0x4a,0xa0, 0x38,0xa0,0x66,0x99,
  0xbc,0xa6,0x47,0xbe, 0x37,0x23,0x1e,0x82, 0x9b,0xcb,0xc9,0x72, 0xac,0x8b,0xdc,0xf9,
  0x2a,0x2a,0x39,0x0a, 0xfb,0x5c,0xed,0x94, 0xe5,0x6f,0x8e,0x8c, 0xd4,0xad,0x1e,0x92,
  0xb1,0x66,0x10,0x55, 0x5e,0x08,0xb1,0xdf, 0x5c,0x89,0x6b,0x36, 0x90,0x77,0x7c,0x31,
  0x9b,0x5c,0x27,0x92, 0x1d,0x6d,0x2a,0x0a, 0x2a,0x2b,0xd9,0xef, 0x46,0xec,0xea,0xf9,
  0x55,0xc7,0xc2,0xd1, 0x3f,0x01,0x85,0x39, 0x8d,0x35,0xab,0x61, 0xff,0xc5,0xdc,0x91,
  0x4e,0x14,0x24,0x05, 0x3f,0x11,0xc4,0x00,                                          
  // 0x1ab8 = 6840 Bytes
};

//
///////////////////////////////////////////////////////////////////////////////

static const u8 ctcode_brres_base_bz2[] =
{
  0x00,0x50,0x2b,0x00, 0x42,0x5a,0x68,0x32, 0x31,0x41,0x59,0x26, 0x53,0x59,0x75,0x3e,
  0x0c,0x2a,0x00,0xc3, 0x65,0xff,0xff,0xff, 0xff,0x57,0xf7,0xf7, 0x71,0x6f,0xfd,0xff,
  0x9a,0x56,0xf7,0x5e, 0x5f,0xbd,0xed,0x7f, 0x4e,0x47,0xd5,0xf7, 0x53,0x7e,0xf7,0xc9,
  0x45,0x3f,0xf5,0xc0, 0x9e,0x94,0x89,0xe0, 0x1c,0x4f,0x7c,0x48, 0xd3,0x41,0xa6,0x81,
  0x52,0x05,0x2a,0x88, 0xaa,0xa5,0x5b,0x58, 0xb6,0xb5,0x5b,0x69, 0x98,0xcd,0xa2,0x83,
  0x56,0x61,0x4d,0x5b, 0x00,0xb5,0x90,0xd8, 0xd0,0x51,0x55,0x54, 0x0b,0x5a,0x92,0x00,
  0x02,0x85,0x09,0x00, 0x34,0x18,0x45,0x2a, 0xac,0x24,0x88,0x26, 0x89,0x8a,0x46,0xa7,
  0xa9,0xe5,0x1e,0xa1, 0xa3,0xd3,0x0a,0x36, 0xa6,0xd2,0x7a,0x80, 0x03,0x4f,0x50,0x01,
  0xea,0x00,0x00,0x68, 0x32,0x03,0x40,0x34, 0x19,0x3d,0x4d,0x00, 0xf5,0x00,0xda,0x83,
  0xd4,0x34,0x69,0xa0, 0x07,0x94,0x06,0x44, 0x01,0xa6,0x86,0x9a, 0x32,0x03,0x46,0x81,
  0x93,0x26,0x8d,0x32, 0x01,0x82,0x32,0x68, 0x34,0xc9,0x82,0x34, 0x31,0x18,0x26,0x9a,
  0x01,0x88,0xd0,0x1a, 0x34,0xc4,0x34,0xd3, 0x26,0x13,0x4d,0x03, 0x46,0x09,0xa3,0x08,
  0xc4,0xc9,0x88,0x03, 0x4d,0x0d,0x34,0x64, 0x06,0x8d,0x03,0x26, 0x4d,0x1a,0x64,0x03,
  0x04,0x64,0xd0,0x69, 0x93,0x04,0x68,0x62, 0x30,0x4d,0x34,0x03, 0x11,0xa0,0x34,0x69,
  0x88,0x69,0xa6,0x4c, 0x26,0x9a,0x06,0x8c, 0x13,0x46,0x11,0x89, 0x93,0x10,0x06,0x9a,
  0x1a,0x68,0xc8,0x0d, 0x1a,0x06,0x4c,0x9a, 0x34,0xc8,0x06,0x08, 0xc9,0xa0,0xd3,0x26,
  // 0x100:
  0x08,0xd0,0xc4,0x60, 0x9a,0x68,0x06,0x23, 0x40,0x68,0xd3,0x10, 0xd3,0x4c,0x98,0x4d,
  0x34,0x0d,0x18,0x26, 0x8c,0x23,0x13,0x26, 0x08,0x91,0x04,0x92, 0x64,0x99,0x26,0x29,
  0xf8,0x1a,0x51,0xa1, 0xa8,0xf2,0x8c,0x8d, 0x33,0x44,0x64,0xc2, 0x64,0x64,0xd0,0x03,
  0x13,0x43,0x6a,0x7a, 0x8d,0xa8,0x68,0xc9, 0xea,0x0d,0x0d,0xa2, 0x18,0x04,0x64,0x03,
  0x47,0xa4,0xf5,0x19, 0x06,0x8d,0x3d,0x46, 0x98,0x98,0x0a,0x89, 0x10,0x84,0x40,0x9a,
  0x09,0x3d,0x3d,0x29, 0x3c,0x4d,0x26,0x9a, 0x7a,0x9e,0x90,0x7a, 0x80,0x7a,0x80,0x7a,
  0x80,0x69,0xa7,0xa2, 0x00,0x68,0x00,0x34, 0x00,0x01,0xa6,0x80, 0x1a,0x34,0x00,0x00,
  0x19,0x3d,0x46,0x80, 0xf2,0x9e,0xa6,0xcf, 0x5f,0x75,0xcf,0x6e, 0x71,0x78,0x5d,0xc7,
  0x69,0x3f,0xdf,0x49, 0x81,0x66,0x30,0x06, 0xab,0x33,0x03,0xab, 0x26,0x04,0x88,0xa7,
  0x4d,0x36,0x00,0xd6, 0xe2,0xdd,0x84,0x02, 0x24,0x66,0x02,0x3a, 0x88,0x6b,0x8c,0x81,
  0x67,0x11,0xa6,0x26, 0x98,0xdd,0x34,0xbc, 0x59,0x1b,0xd8,0xc9, 0x07,0x12,0xc3,0x81,
  0x98,0x28,0xa2,0x71, 0x35,0x33,0x26,0xd3, 0x90,0xb3,0x43,0x41, 0xad,0xcd,0x8d,0x59,
  0xc4,0xc1,0xab,0x37, 0xac,0xd0,0xe6,0x51, 0x6d,0xe7,0x6d,0x48, 0x67,0x2e,0x27,0x1b,
  0x19,0xad,0x97,0x86, 0x24,0x84,0xbc,0x4c, 0x9c,0x0d,0x72,0x6c, 0x60,0x17,0x0e,0x6f,
  0x1e,0x47,0xb3,0x7c, 0xbf,0x6d,0x38,0x4d, 0x30,0xf9,0x3e,0x77, 0x35,0xf3,0x40,0x90,
  0x69,0xa4,0x33,0x3a, 0x98,0x00,0xea,0x19, 0xc2,0x67,0x14,0xf3, 0xc7,0xc1,0x0c,0xe0,
  // 0x200:
  0x0b,0x35,0x92,0xa6, 0xf6,0xd6,0x5c,0xcc, 0xc6,0x09,0x61,0x76, 0xcc,0xb6,0xdc,0x1a,
  0x1a,0x6b,0x7b,0x97, 0x90,0xe6,0xdd,0x23, 0x35,0x6e,0xe7,0x3a, 0xd6,0xc6,0xa8,0xe6,
  0x5e,0x02,0xdd,0xf0, 0x19,0xae,0x15,0x39, 0xb2,0x44,0x30,0x50, 0x94,0xa2,0x94,0x92,
  0x47,0x14,0x93,0x81, 0xcb,0x59,0x34,0x12, 0xf1,0x30,0x8f,0x1d, 0x6e,0x59,0xa9,0xb0,
  0x94,0xd4,0x2c,0xac, 0x19,0x06,0x7e,0x0d, 0x70,0x30,0x71,0x37, 0x14,0x48,0x56,0x00,
  0xc9,0x24,0x16,0x32, 0x11,0x7a,0x60,0x3d, 0x3e,0x87,0x17,0x19, 0x34,0x79,0x0f,0x21,
  0x78,0xea,0x71,0x39, 0x3c,0x1f,0x1b,0xd3, 0x27,0x70,0x7a,0xaa, 0x73,0x2c,0x76,0x2b,
  0xab,0x21,0x92,0x92, 0x32,0xa1,0x36,0x3c, 0x87,0x7d,0x38,0xa4, 0xef,0x33,0x41,0x1e,
  0x04,0x71,0x90,0x0e, 0x37,0x97,0x4e,0x40, 0xe1,0xcd,0x01,0x98, 0x01,0x9d,0x47,0x52,
  0xdc,0xc3,0x3a,0xd6, 0x9a,0xbb,0xdf,0x3a, 0x93,0x2e,0xec,0x9a, 0x96,0x48,0x6a,0xc3,
  0x70,0xf1,0xe9,0xa3, 0xbb,0x61,0xe9,0x39, 0xe4,0x68,0x6b,0xc6, 0x0f,0x02,0x92,0x47,
  0x34,0xb3,0xa1,0x9c, 0x8b,0x3a,0x53,0xbc, 0xf4,0x8c,0x86,0xee, 0xa8,0xe4,0xbe,0x52,
  0xcb,0x44,0x73,0xa4, 0x78,0x8e,0x05,0x93, 0x81,0x7c,0xef,0x30, 0xe2,0x13,0x18,0x78,
  0x14,0x78,0x1a,0xea, 0x6b,0x79,0xc3,0x4d, 0xb2,0x26,0x6b,0xcf, 0x33,0x4c,0x9d,0xf2,
  0x4c,0xd7,0x19,0x15, 0x6e,0xa9,0xe7,0xce, 0xe3,0x93,0x9f,0x2e, 0xa6,0x6e,0x5e,0x73,
  0x42,0x53,0xc8,0xb2, 0x51,0x08,0xba,0xad, 0x3a,0xb1,0xe2,0x82, 0x2a,0x10,0x0a,0xe4,
  // 0x300:
  0x44,0x7a,0x55,0x65, 0x05,0x35,0x19,0xc6, 0xca,0xa6,0x95,0x02, 0xad,0xef,0x66,0xca,
  0x72,0xd6,0xaf,0x1b, 0x1b,0xd9,0xbb,0x2d, 0x66,0xb0,0xb0,0x96, 0x35,0x7c,0x52,0xd1,
  0xa0,0xa9,0xc7,0xa6, 0xe1,0xd5,0x6e,0x29, 0xec,0xed,0x92,0x73, 0x64,0xdc,0x38,0xc3,
  0x8b,0x1a,0x40,0x64, 0x73,0x40,0xa8,0xa3, 0xb0,0xd8,0x52,0xb0, 0x24,0x08,0x0a,0xea,
  0x2f,0x11,0x5d,0x53, 0x79,0x8d,0xeb,0x5a, 0xcb,0xb3,0x5e,0x78, 0x6a,0xb5,0xc6,0xb4,
  0x84,0xd1,0x4a,0x50, 0x92,0x38,0x30,0xc5, 0x15,0xcc,0xe7,0x58, 0x65,0xd2,0xcc,0x5d,
  0x59,0xc6,0xf2,0xea, 0x92,0x29,0x4a,0x52, 0x42,0x47,0x91,0x49, 0x21,0x3a,0x28,0x48,
  0x24,0x4d,0x14,0xf5, 0x14,0x48,0x91,0x18, 0x61,0xdd,0x21,0x13, 0xda,0xf9,0x07,0xe7,
  0x75,0x21,0xee,0xfc, 0x1d,0x6a,0x12,0x42, 0x78,0x2d,0xb6,0x7f, 0xc0,0x71,0xf4,0xf7,
  0x02,0x2d,0x13,0xef, 0xac,0x13,0xe3,0xfe, 0xfe,0x21,0x12,0x75, 0x4f,0x47,0xab,0x0c,
  0x29,0x24,0xf8,0xac, 0x44,0xff,0xb7,0xdd, 0xe1,0x34,0x2c,0x42, 0x39,0xf8,0x3a,0xd5,
  0xdd,0xd8,0x29,0xd7, 0x80,0x8a,0x74,0x40, 0x14,0x90,0x47,0xf7, 0x22,0xa8,0x2a,0xa8,
  0x76,0x23,0xe4,0xb1, 0x30,0x58,0x84,0x99, 0xf1,0xe0,0xf1,0x7d, 0xd9,0x61,0xc6,0xa0,
  0x6a,0x58,0x89,0x00, 0xec,0xce,0xfa,0x75, 0x21,0x44,0x15,0x47, 0x9c,0xa7,0xb5,0x68,
  0x52,0x04,0x15,0x53, 0xb6,0xd3,0x5d,0xf5, 0x05,0x54,0x55,0x0e, 0xa3,0x3b,0x8c,0x40,
  0x60,0x77,0x69,0x07, 0x31,0x08,0x11,0x45, 0xed,0x72,0xa0,0x37, 0x4d,0xe5,0x88,0xfc,
  // 0x400:
  0x4d,0x21,0x3c,0x06, 0x83,0x06,0x4b,0x08, 0xe2,0x9a,0x96,0x58, 0xa8,0x5a,0x91,0x2a,
  0x28,0xb2,0xc5,0xcc, 0xc9,0x62,0xd9,0x65, 0x82,0xdb,0x29,0x14, 0x59,0x17,0xcb,0x24,
  0x30,0x54,0x96,0xd1, 0x49,0xa9,0x72,0x8a, 0xa4,0x65,0x6e,0x6b, 0x2c,0x8b,0x24,0xe5,
  0x8f,0x2d,0x48,0xcd, 0x4c,0x8c,0x2c,0x3c, 0xfa,0xd4,0x34,0x2c, 0xa4,0x96,0x50,0xa1,
  0x62,0x16,0x50,0xb0, 0xb1,0x61,0x64,0xb1, 0x64,0x9a,0xdc,0xd0, 0xd0,0xe6,0x64,0xca,
  0x52,0x52,0x58,0x2c, 0xa3,0x9a,0x26,0x05, 0xdb,0x4d,0x2a,0x80, 0x2a,0x86,0x59,0x4e,
  0x9a,0x40,0x0d,0xb0, 0x5d,0xc4,0xa9,0xa9, 0xb8,0xc3,0xa1,0x79, 0xbd,0x5e,0x85,0x44,
  0xea,0x58,0xa2,0xc3, 0xab,0x12,0x8b,0xd5, 0xe2,0x54,0x84,0xe6, 0x84,0x9b,0xeb,0x21,
  0xbd,0xb8,0x1b,0xda, 0x56,0xc6,0x4e,0x7a, 0xe7,0x54,0x71,0xb3, 0x7a,0x02,0x68,0x0c,
  0xd6,0xa4,0xc8,0x66, 0xb7,0x42,0x56,0x52, 0xb7,0x1d,0x81,0xac, 0x16,0xe2,0x9b,0x65,
  0x91,0x24,0xcb,0x27, 0x34,0xc9,0x60,0x23, 0x81,0xc8,0xd6,0xc3, 0x36,0x1d,0x4d,0xe8,
  0x5b,0x50,0x18,0xe9, 0x86,0x2c,0xd5,0xb5, 0x04,0x44,0x1f,0x13, 0xa6,0x95,0x05,0xfb,
  0xc8,0x81,0x7d,0xec, 0x1f,0x73,0x51,0xf4, 0x35,0xeb,0xd4,0x5b, 0x27,0xd4,0x2a,0x0f,
  0xa5,0x70,0xf5,0xe2, 0x44,0x8f,0x45,0x3e, 0x1a,0x13,0x55,0x23, 0xf6,0xf6,0x20,0x99,
  0x64,0x91,0xfc,0x4c, 0xf5,0x8f,0x5a,0x6e, 0x57,0x41,0x98,0xb4, 0xf6,0x7b,0x5e,0xd2,
  0xcd,0x6d,0xc4,0xb1, 0xc5,0xe6,0xc6,0xf9, 0xc9,0x81,0x64,0x16, 0x96,0xb5,0x2c,0x87,
  // 0x500:
  0xb3,0x17,0xb1,0x2a, 0x22,0xf2,0x80,0x9a, 0x9f,0x1a,0x23,0x71, 0x10,0x91,0x02,0x41,
  0x10,0xec,0xb1,0x39, 0xa2,0xa1,0x71,0x5b, 0x8a,0x0f,0x9d,0x04, 0x3b,0x92,0x3d,0xca,
  0x05,0xcc,0xd4,0xf0, 0xe1,0x88,0x97,0x8a, 0x7c,0x28,0x09,0x5d, 0xaf,0xbe,0xee,0x76,
  0xeb,0xd4,0x1f,0x7e, 0x77,0x78,0x10,0x4e, 0xbc,0x11,0x0e,0xfe, 0x22,0x77,0x60,0x81,
  0xdf,0xc6,0x3d,0x52, 0xc8,0x30,0xf9,0x66, 0x24,0x1a,0x0a,0x44, 0x3d,0xe9,0x75,0x46,
  0x59,0x40,0x80,0x2d, 0xc0,0x50,0x05,0x50, 0xd3,0x04,0x94,0xc6, 0xf2,0xde,0x1c,0x66,
  0x62,0xc1,0x18,0x16, 0x12,0x4f,0x82,0xc8, 0x81,0x92,0xb5,0xc3, 0x5a,0x49,0xaa,0x89,
  0x65,0x8b,0x41,0xc1, 0x4d,0x6e,0x03,0x90, 0x20,0xea,0x22,0x97, 0x15,0x1c,0x41,0x55,
  0x24,0x05,0x09,0xc9, 0x50,0xe9,0xd0,0x64, 0x3a,0x66,0x10,0xcd, 0xa5,0xac,0x09,0x39,
  0xea,0x47,0x1a,0x55, 0xcc,0x36,0x45,0x3a, 0xb1,0x03,0xc2,0x0e, 0xb8,0x57,0xe6,0x45,
  0x04,0x47,0xe6,0xc5, 0x10,0x70,0x7a,0x1f, 0xda,0xed,0xbb,0x09, 0xd6,0xcb,0xd8,0x02,
  0x1b,0xc0,0xc2,0x3d, 0xc2,0xb1,0x3e,0x98, 0x79,0xe1,0x90,0x2f, 0x44,0x9e,0xe2,0xdf,
  0x43,0xc7,0x20,0x79, 0xb7,0x67,0xda,0x76, 0x7e,0x6f,0xbf,0x6c, 0x75,0x68,0xce,0xb8,
  0x5f,0x44,0xbc,0x87, 0xb2,0x6e,0xfe,0x9f, 0x63,0x4e,0x5d,0x00, 0xb1,0xe6,0xc2,0x21,
  0xa8,0xf2,0x8a,0xf9, 0x30,0x01,0x4e,0x7e, 0x6f,0xd6,0xbe,0x4c, 0xe5,0x5c,0xd5,0x1d,
  0x3d,0x7d,0xeb,0x30, 0xa5,0x16,0x68,0xdf, 0x1a,0x9f,0xa2,0xff, 0x17,0xdf,0xfc,0xff,
  // 0x600:
  0x51,0xdb,0xb7,0x78, 0x3d,0x55,0xdf,0x8c, 0x0c,0x0f,0x3d,0x69, 0x20,0xd3,0x61,0x02,
  0x0c,0xf6,0x11,0x7a, 0xa4,0xee,0xa8,0x32, 0xb0,0xd7,0x7c,0xbb, 0xb0,0x31,0x70,0x8b,
  0xc1,0xac,0x56,0x60, 0x89,0x6a,0xe3,0xca, 0x7a,0x2d,0xc4,0x8f, 0x1a,0x7f,0x7e,0xc0,
  0xb0,0x80,0x19,0x68, 0x0c,0x85,0xc2,0x3d, 0x97,0xf3,0x7e,0x98, 0x7e,0x08,0x63,0x0e,
  0xc7,0xab,0x76,0xc8, 0x04,0x66,0xe6,0xae, 0x56,0x09,0x27,0x50, 0x01,0xd7,0xa3,0x06,
  0x68,0xc3,0x1b,0x36, 0x40,0x68,0xe7,0x2a, 0x89,0x0b,0x6a,0x75, 0xf8,0x78,0x3b,0x77,
  0xfa,0x23,0xbe,0xfc, 0x9a,0x0e,0xd2,0x3b, 0x06,0x7d,0xfd,0x9d, 0x1c,0x8b,0x37,0xc6,
  0xf5,0xa7,0xde,0xd7, 0xa1,0x4d,0xdd,0xad, 0x9e,0x2e,0xab,0x31, 0x99,0x9a,0xfb,0xdb,
  0xab,0xaf,0x18,0x73, 0x67,0x3d,0xc7,0xa3, 0x65,0xb9,0x93,0x47, 0x3a,0xab,0x51,0xf6,
  0xdd,0x06,0x4d,0xdf, 0xd2,0x5f,0xad,0xef, 0x60,0xea,0x71,0xe2, 0x7d,0x5a,0x68,0xec,
  0x6e,0x79,0x8c,0xac, 0x73,0x00,0x51,0x55, 0x15,0xab,0xf5,0xfe, 0xb2,0xad,0x56,0xd3,
  0x55,0x25,0x0c,0x26, 0x4e,0xcd,0xe4,0x40, 0xb7,0x19,0x80,0x26, 0x4d,0x50,0x0c,0xbd,
  0x0d,0x74,0x5d,0xe0, 0x34,0xf2,0xc6,0x26, 0x54,0xb7,0x26,0x8c, 0x21,0x9d,0x3e,0x03,
  0x7d,0x47,0x28,0x60, 0x0e,0x03,0x91,0xb1, 0x83,0x8e,0xd2,0x13, 0x70,0x97,0x56,0x5b,
  0x85,0xe1,0xda,0x6b, 0xa9,0xc8,0xe5,0x86, 0x8d,0x56,0xc4,0xad, 0xd4,0xd1,0x66,0x5e,
  0x31,0xa8,0x19,0x03, 0x00,0x3f,0x0b,0xc5, 0x03,0xb7,0xe3,0x77, 0x40,0xee,0x88,0x6d,
  // 0x700:
  0x18,0x43,0x26,0x8b, 0x24,0x0e,0xcb,0xf6, 0xa6,0xf3,0x34,0x08, 0x9c,0xbb,0x47,0x7f,
  0x4d,0xce,0x4a,0xc1, 0x72,0x8b,0x6d,0x2a, 0x8d,0x11,0x64,0x26, 0xd9,0xe3,0x31,0xae,
  0x58,0x5a,0x61,0x32, 0xd7,0xe9,0xef,0x56, 0x80,0xfd,0x68,0xfa, 0xf8,0xad,0xf9,0x35,
  0xd8,0xb4,0xa0,0x22, 0xd3,0x00,0x61,0x67, 0x01,0xb1,0x00,0x75, 0x1e,0x0b,0x01,0xed,
  0xa6,0x02,0x72,0x1b, 0x60,0x01,0xd8,0xab, 0xfa,0x29,0xec,0x38, 0xfd,0x4b,0x5c,0x40,
  0x84,0x53,0x4d,0x3a, 0x6c,0xf1,0xe0,0x07, 0x98,0x04,0x10,0x32, 0xe4,0x08,0x9c,0xff,
  0xd1,0xe4,0x79,0xbd, 0x9d,0x0e,0xa8,0xf8, 0xc7,0x5e,0xcd,0x77, 0xd7,0x01,0xef,0x8d,
  0xf8,0xfc,0x23,0x3e, 0x98,0xea,0x72,0x27, 0xab,0x66,0x4e,0xeb, 0x87,0x0d,0x08,0x45,
  0x64,0xb6,0xf1,0x84, 0x23,0x16,0xba,0x80, 0xe1,0x87,0xe6,0x14, 0x9e,0xbb,0x04,0xc2,
  0x68,0x1d,0x35,0x80, 0x16,0xd3,0x6b,0x3e, 0x10,0x15,0x80,0x32, 0xd1,0x8e,0x6e,0xe3,
  0xbf,0x03,0x39,0x00, 0xe9,0xfe,0x2b,0x40, 0xb8,0xab,0xe1,0xc5, 0xec,0x7e,0x76,0xdb,
  0x7e,0x00,0xfe,0x2f, 0x09,0x6a,0x84,0x21, 0xfe,0x4a,0x1a,0xce, 0x3a,0x9c,0xd9,0x4d,
  0xd3,0x38,0x13,0x55, 0x46,0x2e,0x34,0x83, 0x98,0xdc,0x33,0x33, 0x33,0x57,0xe3,0xe2,
  0x48,0x99,0x44,0x49, 0x41,0x21,0xd8,0xb2, 0x2e,0xf7,0xde,0x5a, 0xdd,0x3e,0x53,0xa3,
  0x38,0x30,0x82,0x8a, 0x76,0x04,0xb3,0xbf, 0x14,0xe2,0x78,0x1e, 0x47,0x3c,0x4d,0x4d,
  0x1b,0x64,0x19,0x93, 0x2b,0x2a,0xe4,0xcb, 0x43,0x5a,0x99,0xc6, 0xe1,0xd0,0xa1,0x86,
  // 0x800:
  0x52,0xef,0x53,0xa0, 0xcc,0xd4,0xe3,0x89, 0xae,0x1e,0xbe,0x6b, 0x7c,0x21,0x90,0x26,
  0xf4,0x89,0xa0,0xce, 0x8e,0x0f,0x9f,0xe2, 0xfa,0x7e,0xd9,0xe1, 0x21,0x99,0x80,0xed,
  0xdd,0xc6,0x01,0xde, 0xfc,0x2a,0x0f,0x97, 0xf0,0x73,0x70,0x52, 0x20,0x46,0x78,0x94,
  0x9e,0x50,0x11,0xb4, 0x0f,0x53,0x1c,0x4b, 0x8f,0x4b,0x47,0x35, 0x9e,0xfb,0xfa,0xbb,
  0xd6,0xe7,0xa2,0x34, 0x81,0x29,0xd5,0x7d, 0xdc,0x1e,0x5d,0xbc, 0xe8,0x0b,0x7b,0x94,
  0x04,0x25,0x39,0x5a, 0xb6,0x0f,0xa4,0x03, 0x6a,0xc1,0x14,0xc1, 0x35,0x05,0x89,0x2c,
  0xc9,0xb8,0x06,0x47, 0xbe,0x17,0x2a,0xe4, 0x6b,0x36,0x98,0x0c, 0x84,0x55,0x6c,0x9c,
  0x03,0xf4,0x0f,0x24, 0x1e,0x46,0x83,0x9f, 0xec,0x13,0xb2,0x46, 0x46,0x48,0x19,0x49,
  0x99,0x66,0x2c,0xb4, 0xc0,0x66,0x29,0x04, 0x80,0x90,0x64,0x16, 0x68,0x68,0x7e,0x46,
  0x6f,0x99,0xbc,0xe0, 0x2f,0xdc,0x63,0x41, 0xc8,0xe4,0xe2,0x75, 0x8c,0x86,0xf3,0xb6,
  0x1a,0xb9,0xa7,0x36, 0x92,0xf5,0x2e,0xa8, 0xad,0x49,0x21,0x28, 0x31,0xad,0x00,0x69,
  0x8d,0x37,0x70,0xd8, 0xe8,0x6c,0x6e,0x8d, 0x1f,0x28,0xeb,0x56, 0x5a,0xed,0xd5,0xc4,
  0x35,0x38,0x2e,0xd7, 0x76,0x5d,0x9a,0xd6, 0xbc,0x68,0xd4,0x3b, 0x2f,0x83,0x9e,0xf3,
  0x80,0xbe,0x34,0xc3, 0x91,0xc9,0xba,0x9d, 0xcb,0x38,0xee,0x66, 0xfa,0x99,0x59,0xa9,
  0xc0,0xe2,0x37,0x85, 0x1c,0x4d,0x74,0x9a, 0xe8,0x6c,0x32,0x05, 0x01,0x76,0xe6,0xda,
  0x00,0xc4,0xb8,0x3b, 0xce,0x94,0xc6,0x1d, 0xb1,0xd0,0x06,0xdc, 0x30,0x1d,0x4b,0xb1,
  // 0x900:
  0xaa,0x38,0xe6,0x72, 0x75,0xb9,0x14,0x5a, 0x35,0x91,0xc7,0x50, 0xb1,0x9a,0x6e,0x6c,
  0x73,0xbe,0x46,0x66, 0x75,0xb2,0x8d,0x86, 0x59,0xbb,0xcc,0xc1, 0xdb,0xb4,0xec,0x7b,
  0x75,0xe2,0x3f,0x59, 0x28,0xf9,0x9e,0x8f, 0x70,0xfe,0xef,0x33, 0x88,0x3a,0x3e,0x03,
  0x0c,0x32,0x7d,0x03, 0xbf,0x03,0x3b,0x0a, 0x39,0x0c,0xa3,0x9d, 0x33,0x57,0x5b,0x3e,
  0x6f,0xbb,0x27,0xd8, 0xe1,0xc0,0xa8,0xef, 0xb8,0xfb,0xab,0x0d, 0x67,0xcd,0x1a,0xef,
  0xb7,0xce,0xfc,0xbe, 0xb5,0x17,0xfa,0xd3, 0xe9,0x73,0xf5,0x27, 0x09,0x39,0x14,0x72,
  0xb2,0xd6,0xe8,0xec, 0x79,0x1e,0x42,0x00, 0x65,0x84,0x80,0xa6, 0xec,0x07,0x41,0x45,
  0x50,0x13,0xcc,0x8a, 0xe3,0x18,0x58,0x7a, 0x8e,0x70,0x3d,0x8a, 0x80,0xb5,0x58,0xed,
  0x3f,0x0a,0x67,0xc2, 0x79,0xa6,0x5a,0x76, 0x7a,0xa5,0x64,0xcb, 0xd4,0xd6,0x19,0x3e,
  0x2f,0xbc,0x16,0x7e, 0x29,0xf8,0xc6,0x72, 0x1f,0x3e,0x13,0x9c, 0x98,0x30,0x8a,0xcb,
  0xdd,0x38,0x96,0x30, 0x03,0x9e,0x7a,0x85, 0x19,0x23,0x46,0x2a, 0x65,0xd1,0x15,0xe0,
  0xe1,0xe9,0x9d,0x9d, 0xad,0x86,0xd8,0xc0, 0x36,0x9e,0x75,0x87, 0xd9,0xf5,0x5f,0x95,
  0x3c,0xd0,0x3c,0xa5, 0x70,0xe2,0x57,0xb8, 0x70,0xf9,0x76,0x13, 0x59,0x75,0x2b,0x3e,
  0x3e,0xb7,0x7b,0xb0, 0x37,0x6c,0x63,0x58, 0x37,0x48,0x14,0x4a, 0x0f,0xff,0xf1,0x74,
  0x07,0x28,0xe6,0x70, 0x3b,0x6b,0xb6,0xfb, 0x71,0x37,0xf5,0x68, 0x71,0x3e,0x7f,0xe3,
  0xfc,0x9d,0x0e,0xbb, 0x9c,0xab,0x4a,0x68, 0x6b,0xd0,0x1b,0x9e, 0x5e,0x5a,0xf4,0x9b,
  // 0xa00:
  0x34,0x69,0xb1,0xa3, 0xc8,0x69,0x86,0x72, 0x9d,0x2f,0x13,0x61, 0x4e,0x59,0x34,0x72,
  0x6e,0x6b,0xb3,0x89, 0xc9,0xc9,0x4d,0xec, 0xe1,0x5d,0x6a,0x71, 0x38,0x9a,0x99,0x99,
  0x86,0xa1,0xd0,0xe6, 0x74,0x30,0x6a,0x76, 0xde,0xe6,0xb2,0xf5, 0x2c,0x87,0x33,0x54,
  0x67,0x68,0xed,0x93, 0x41,0x80,0xaa,0x73, 0xa6,0xb2,0x04,0xcb, 0xa7,0x34,0x86,0x83,
  0x2c,0xcb,0x9a,0x94, 0xf0,0x17,0x80,0x2e, 0x00,0xb0,0x5f,0xb9, 0xee,0x3d,0xe7,0xc7,
  0xf4,0x7d,0xf9,0xaf, 0xb4,0x30,0x9f,0x17, 0x26,0x4a,0xf9,0xc1, 0x63,0xb4,0xd0,0xe6,
  0x58,0xf9,0xdd,0xb2, 0x3f,0x8c,0x72,0x4e, 0xc1,0xa8,0x98,0x3f, 0x07,0x1d,0x6b,0x32,
  0x56,0x51,0x51,0x66, 0x54,0x9d,0x10,0x7a, 0xcb,0x16,0x7b,0xab, 0x56,0xca,0xea,0x02,
  0x23,0x41,0x71,0xdd, 0x70,0x3e,0x05,0xd4, 0x5f,0x5d,0x00,0xd2, 0x5c,0xa2,0xc8,0xbd,
  0x4e,0x5d,0xce,0xae, 0x16,0x58,0x52,0xbb, 0xb5,0xbd,0x23,0x77, 0x9a,0x7a,0xb7,0x72,
  0xcc,0x3b,0x4e,0x26, 0xe5,0xfc,0x50,0xc7, 0x08,0x10,0x0d,0xb5, 0xfa,0xb0,0x14,0x6c,
  0x0c,0x4f,0x0b,0x4f, 0x3e,0xfa,0xaa,0x06, 0x58,0xe0,0x6c,0x40, 0xa4,0x50,0x8a,0xd9,
  0x49,0x94,0xbd,0x5a, 0x12,0x01,0x62,0x46, 0x31,0x3f,0xaf,0xc3, 0xef,0x49,0xd6,0xe9,
  0x7c,0x37,0xd1,0x00, 0x66,0x01,0x1b,0x0d, 0xb2,0x30,0xe2,0xfe, 0x06,0x4e,0x03,0x5c,
  0x73,0xcc,0xd6,0x0d, 0x2f,0x1f,0x3b,0xed, 0x8a,0x4e,0xcd,0xa4, 0xe9,0xca,0x3b,0x75,
  0xc0,0x69,0xd4,0xeb, 0x2b,0x7a,0x2e,0x8e, 0x6b,0x3e,0xf3,0x83, 0x61,0x75,0x65,0xee,
  // 0xb00:
  0x77,0x99,0xcd,0xeb, 0x0d,0x35,0xcb,0x2b, 0x9e,0xc3,0x7d,0x3a, 0x1c,0x75,0x28,0x64,
  0xeb,0xa7,0x32,0xf7, 0xe4,0xd6,0xa7,0x02, 0xcd,0xcd,0x06,0xe6, 0xf6,0xa3,0x58,0x59,
  0x66,0xa5,0x91,0xbc, 0xe4,0x39,0x0d,0xa1, 0xa8,0x6b,0x7a,0x1c, 0x51,0x7b,0x71,0x1b,
  0x98,0xd4,0xd9,0xc5, 0x3b,0xa3,0xca,0xd9, 0x3c,0xfe,0x86,0x7b, 0x06,0x98,0x3d,0xc1,
  0xf9,0xfd,0x37,0x45, 0x94,0x3d,0xc7,0xbc, 0x64,0xf9,0x05,0x91, 0xd4,0x7c,0xdf,0xbe,
  0xd1,0xd4,0xd8,0xc7, 0x5f,0x33,0xb4,0xdd, 0x2c,0xfe,0x1f,0x07, 0xca,0xbc,0xd9,0x1d,
  0x41,0x47,0x5e,0x78, 0x3a,0x0f,0x9f,0x3f, 0x83,0x30,0x33,0xeb, 0x6c,0x6a,0x27,0xfa,
  0xe8,0xed,0x43,0xd5, 0xd9,0xf2,0x73,0xe3, 0xce,0x72,0x1a,0xe6, 0x58,0x42,0x53,0x49,
  0x13,0xe5,0xc0,0xeb, 0xc5,0xb0,0x39,0x4e, 0x77,0x97,0x9f,0x90, 0x3c,0xd9,0x9f,0xfa,
  0xc7,0xfe,0x59,0xc4, 0x07,0x33,0x9c,0x02, 0xd5,0x5d,0xd1,0x88, 0x48,0x34,0x99,0x04,
  0x00,0x64,0x5f,0xee, 0x8c,0xbf,0x17,0x03, 0xe9,0x08,0x34,0x3f, 0xe3,0xfa,0x8c,0x63,
  0xb3,0x83,0xba,0x13, 0x1a,0xc0,0x50,0x5c, 0x74,0xff,0x6d,0x38, 0x20,0x1d,0x8d,0x01,
  0xb0,0x2f,0x83,0x8a, 0x23,0x03,0x5f,0xeb, 0xf0,0x7a,0xd5,0xe0, 0xea,0x51,0xce,0xac,
  0xdc,0xb9,0xb9,0x4f, 0x1a,0xef,0x31,0xfd, 0x5f,0x3e,0xdd,0x06, 0xfb,0x09,0xa2,0xe2,
  0x90,0xe5,0xd4,0x64, 0x2c,0x53,0x61,0xda, 0x61,0x96,0x2c,0xb9, 0xa3,0x17,0x82,0xb8,
  0x5b,0x64,0x5e,0xa6, 0xea,0xc0,0xce,0xdb, 0x4b,0xaa,0x43,0x16, 0x05,0x60,0x08,0x6d,
  // 0xc00:
  0xc3,0x75,0x9d,0x39, 0x74,0xc6,0x10,0xcc, 0x72,0x06,0x5c,0x55, 0xa6,0xea,0xb7,0x49,
  0xb9,0x52,0xc4,0xac, 0xd1,0xa9,0xa7,0x3b, 0x98,0x70,0xbd,0x1e, 0x4c,0x03,0xc0,0x83,
  0xd7,0x22,0x07,0xd5, 0x9d,0xec,0x09,0x42, 0x98,0x7c,0x1a,0xc0, 0x04,0x7d,0x13,0x10,
  0xe9,0x87,0xdc,0x4e, 0xff,0xcc,0x03,0x9f, 0x58,0x8f,0x2d,0xbd, 0x96,0xf0,0x48,0xef,
  0xa6,0x8d,0x38,0x72, 0xf3,0xf7,0x72,0x1d, 0x11,0x33,0x14,0x59, 0xec,0x52,0xe7,0x52,
  0xa7,0x06,0xab,0x6d, 0x09,0xc9,0x41,0x9d, 0x41,0x0b,0x59,0x87, 0x56,0xb5,0x8e,0xe3,
  0x28,0x3b,0xb4,0xc2, 0xb8,0x7a,0xf7,0x3d, 0x27,0x79,0xe7,0x72, 0x76,0x91,0xfa,0x10,
  0xe2,0x6f,0xcf,0x21, 0x36,0x2c,0x9e,0x59, 0x30,0xf1,0x5a,0xf2, 0x9b,0x8d,0x78,0x94,
  0x3f,0xa2,0x16,0x0d, 0x4e,0x26,0x67,0xb2, 0xc8,0xc9,0x0f,0x37, 0x9e,0x78,0x9e,0x5f,
  0x53,0xf0,0xeb,0xfe, 0xaf,0xad,0x79,0x73, 0x4f,0x65,0xec,0xd6, 0x87,0xa7,0xd5,0x9c,
  0x21,0xe2,0x51,0xab, 0xeb,0xf3,0xde,0x7a, 0xb6,0x1c,0x4d,0xeb, 0xd3,0x22,0xf1,0x20,
  0x3c,0xbe,0xbf,0x58, 0x40,0xe9,0x62,0x07, 0x75,0x9d,0x30,0x13, 0xbd,0x17,0xbd,0x01,
  0x1f,0x5b,0x30,0x04, 0x43,0x8e,0x06,0xed, 0x9a,0x82,0x26,0xac, 0x83,0x5a,0x99,0xb2,
  0xa8,0x5f,0x98,0x2c, 0x99,0x34,0x32,0x5b, 0x0d,0x0a,0x8c,0xb1, 0x08,0x9c,0x8a,0x41,
  0xc4,0xb2,0x1a,0x14, 0x23,0x2a,0x06,0x59, 0x19,0x50,0x8d,0x58, 0x1c,0xd8,0x91,0x97,
  0x5c,0x64,0x91,0x1c, 0x8a,0x41,0x37,0x63, 0x46,0x60,0xdd,0x96, 0xa7,0x4b,0x24,0xb6,
  // 0xd00:
  0x23,0xa9,0x40,0x51, 0x44,0x70,0x2c,0x81, 0xc0,0xb2,0x47,0x13, 0x7d,0x4e,0x28,0xc0,
  0x31,0x8c,0x01,0x48, 0x88,0xe6,0x2a,0x8b, 0x00,0xcd,0x0a,0xac, 0x2c,0x28,0x56,0xda,
  0x4e,0x2c,0x4b,0x2d, 0xb1,0x65,0x23,0x82, 0x91,0x31,0x61,0x6c, 0x25,0xa9,0x12,0x4e,
  0x02,0xc8,0x38,0x28, 0x8e,0xae,0x8b,0x24, 0x4c,0x29,0x24,0xc5, 0x88,0xe4,0x51,0x24,
  0xb6,0xd4,0x91,0x2d, 0x10,0xb4,0x92,0x72, 0x59,0xc0,0xa0,0xd9, 0x6d,0x08,0x97,0x14,
  0xc9,0x23,0x05,0x44, 0xd0,0x68,0x36,0x01, 0x6a,0x0d,0xb1,0x45, 0x96,0x44,0xb6,0x24,
  0x4c,0x54,0x9c,0xa9, 0x75,0x9a,0x0e,0xb2, 0x24,0x0d,0x9a,0x68, 0x55,0x2c,0x02,0x28,
  0x29,0xa8,0xa2,0xa1, 0x18,0x29,0x20,0x2a, 0x03,0x92,0x88,0x92, 0x72,0x2c,0x07,0x05,
  0x89,0x26,0x54,0x92, 0x75,0x44,0x37,0x60, 0x5a,0x02,0x0a,0x2b, 0x8d,0xd2,0x2e,0xe0,
  0x02,0x2e,0x73,0x4a, 0x53,0x05,0xcc,0x04, 0x48,0x6a,0xc8,0x87, 0x54,0x89,0xb1,0x52,
  0x13,0x43,0x43,0x36, 0x2c,0x80,0x76,0x04, 0x50,0xa8,0xa0,0xe6, 0x22,0xe4,0x82,0x2c,
  0x02,0x2a,0x03,0x6c, 0x40,0x36,0x39,0x1d, 0x6a,0x07,0x05,0x84, 0x96,0xc2,0xdd,0x28,
  0x2d,0x43,0x62,0xa3, 0x9a,0x48,0x73,0x50, 0x9b,0xa4,0x8d,0x50, 0x93,0x62,0xc4,0x90,
  0x8c,0x15,0x00,0x96, 0xa2,0x49,0x65,0x24, 0x35,0x62,0x43,0x2c, 0x44,0x9d,0x0a,0xd5,
  0x44,0x83,0xd8,0x3c, 0x89,0x65,0x7b,0x2c, 0xf3,0x4b,0x3d,0x1e, 0xd9,0x9c,0x5f,0x54,
  0x30,0x33,0x26,0x64, 0x7b,0x2c,0xd5,0x9f, 0x6b,0x77,0x4b,0xac, 0x47,0x69,0x72,0x1e,
  // 0xe00:
  0xfb,0xdb,0xa1,0xa9, 0x5d,0x8a,0x7d,0x7f, 0x3e,0xff,0x99,0xf4, 0x71,0x3b,0xdf,0x3b,
  0x07,0xc3,0xb7,0x80, 0xd4,0xf4,0x0e,0x70, 0x30,0x7a,0x71,0xe8, 0xb2,0x38,0x1b,0xae,
  0xab,0xd5,0xdd,0xb9, 0x82,0xf2,0x79,0x9c, 0xb9,0xb8,0xe4,0x0e, 0x4c,0x00,0xae,0x59,
  0x0c,0x6d,0x8e,0x40, 0x36,0x04,0x00,0xa4, 0x9e,0xa8,0xb8,0x12, 0xbb,0x89,0xde,0x45,
  0x2b,0x18,0x02,0x49, 0x92,0x2d,0xd9,0x00, 0x5c,0xad,0xbe,0xe5, 0x3b,0x80,0xe8,0x6e,
  0x65,0xc0,0x05,0x72, 0x6a,0x83,0x74,0x6f, 0x9b,0x21,0x6f,0x34, 0x1e,0x23,0x39,0x01,
  0xc0,0x1a,0x8f,0x89, 0x93,0xa8,0x40,0x92, 0x43,0xe9,0xfa,0xa7, 0xd9,0xf9,0xfd,0x7b,
  0xd3,0xaf,0x2f,0xb8, 0xbd,0xe6,0x99,0xe7, 0x84,0x8d,0x25,0xf4, 0x53,0x22,0xe1,0xa6,
  0x80,0xc3,0x03,0xbe, 0x8e,0x06,0x47,0x1c, 0x07,0x1d,0x67,0x1d, 0xf6,0xf0,0xde,0xe6,
  0xda,0xd9,0x0c,0x32, 0x4d,0x9b,0x98,0x7d, 0xd5,0x8d,0xf9,0xce, 0x75,0x3b,0x0b,0x05,
  0xf2,0x9d,0x6b,0xd8, 0x1a,0xbd,0x4b,0xb9, 0xc8,0xf3,0xf1,0xd3, 0x9b,0x65,0xe3,0xb6,
  0x25,0xea,0x46,0x74, 0x1a,0x1c,0x8e,0x77, 0x06,0xc6,0x86,0x22, 0x8b,0xba,0x6f,0x38,
  0x5c,0xd3,0xa7,0x5a, 0x02,0x01,0xc1,0xa0, 0x84,0x44,0x4f,0xf8, 0xfa,0xbe,0xab,0xf9,
  0x3a,0x27,0xff,0x7b, 0x90,0x0f,0x8b,0xf5, 0xbe,0x45,0x80,0x7f, 0xef,0x71,0xe8,0x7d,
  0x5c,0x51,0x5e,0xa9, 0x00,0x03,0xd1,0x80, 0x76,0x40,0x69,0x18, 0xc9,0x12,0x8a,0xf8,
  0xa5,0x91,0x1f,0x00, 0x71,0x32,0xc4,0x2c, 0xb2,0x24,0xa4,0x94, 0xa0,0xa9,0x3e,0x1a,
  // 0xf00:
  0x11,0xf0,0x0b,0xa0, 0xf8,0x66,0x48,0x4d, 0x4b,0xf1,0x4d,0x60, 0x99,0xc0,0xcf,0xe6,
  0xff,0x2b,0xff,0xdd, 0x76,0x1c,0x42,0x85, 0x84,0x4b,0x29,0x2a, 0xc2,0xca,0x22,0xd8,
  0x9e,0xe9,0x6c,0xa6, 0x58,0x59,0x62,0xcb, 0x65,0x52,0x55,0xb0, 0xa4,0xa2,0x8b,0xe5,
  0x66,0xee,0x59,0xea, 0x0c,0xf1,0x34,0x78, 0xa7,0x95,0x4f,0x39, 0x99,0x23,0x61,0x65,
  0xf9,0x73,0x24,0x9c, 0xf3,0x30,0xdd,0x88, 0x5b,0x92,0xc8,0x2a, 0x91,0x24,0x35,0x2a,
  0x46,0xe5,0x92,0x64, 0xa1,0xbb,0x32,0x54, 0x39,0xa7,0xa4,0x68, 0x31,0x23,0xc0,0xa0,
  0xed,0x64,0xe9,0x49, 0x1a,0x16,0x6e,0x55, 0x16,0x49,0x42,0xc8, 0x8a,0x16,0x3c,0x4a,
  0x81,0x97,0xe5,0x7d, 0xbf,0x6d,0xcd,0xc2, 0x28,0xae,0x54,0x98, 0x51,0x65,0x16,0x59,
  0x0a,0x39,0xc4,0xed, 0x2f,0x97,0x18,0x1c, 0x51,0x6c,0x8b,0x28, 0x96,0x92,0xd9,0x3c,
  0x8a,0x99,0x64,0x45, 0x54,0x5d,0xe4,0x83, 0x72,0xa4,0x71,0x4d, 0x8b,0x08,0xb2,0x2c,
  0x27,0x80,0xc1,0x89, 0x16,0x96,0xad,0x92, 0x5a,0x2a,0x2c,0x49, 0x3d,0x16,0x66,0xe5,
  0xca,0x1e,0x6c,0x19, 0x2c,0xd5,0x48,0x78, 0xed,0x32,0x06,0xd5, 0x09,0x6c,0x82,0xaa,
  0x6a,0x54,0x98,0x3c, 0x62,0x1b,0xb1,0xa0, 0xd6,0x13,0x32,0x62, 0x37,0x93,0x24,0xd5,
  0x87,0x32,0xf8,0xb1, 0xa8,0x5e,0xe5,0x32, 0x59,0x0e,0xac,0xdd, 0x65,0x1a,0xb1,0x45,
  0x84,0xb6,0x5a,0x5a, 0xc6,0xe8,0x1e,0x4c, 0x00,0xd6,0x5a,0x6a, 0xa0,0x61,0xe5,0x8a,
  0x03,0x00,0x6a,0xd1, 0x7e,0x27,0x6b,0xcf, 0x03,0xa4,0x7e,0x38, 0xfb,0x5f,0xe9,0x8e,
  // 0x1000:
  0xd1,0x27,0xbe,0x2c, 0x83,0xff,0x43,0xf0, 0x06,0x44,0x3d,0xb0, 0x79,0xb0,0x92,0xc0,
  0x9e,0xf5,0x43,0xda, 0x16,0x41,0xa1,0x60, 0xf5,0xd9,0x35,0x2b, 0x25,0x09,0xa0,0xf6,
  0x62,0x24,0xbf,0xee, 0x1f,0xc7,0xd4,0x44, 0xdd,0x96,0xc8,0x2c, 0x25,0x96,0x96,0x8a,
  0x2c,0x25,0x0f,0x80, 0x62,0x65,0x46,0x58, 0xcb,0x2c,0xa1,0xd6, 0x4c,0x93,0xe0,0xc8,
  0xd3,0xc3,0x4a,0x51, 0x7c,0x29,0xe0,0xe5, 0xa6,0xf0,0x86,0x04, 0xa3,0x40,0x69,0xc1,
  0x9b,0xa6,0x3a,0x03, 0x99,0xc0,0x6f,0xc4, 0xb0,0xdc,0xba,0x94, 0xc9,0x0b,0x20,0x43,
  0x9e,0xb2,0x09,0xa9, 0xc6,0x41,0x95,0x27, 0x61,0xbc,0x23,0xa2, 0xc6,0x58,0xe6,0xe6,
  0x61,0xc0,0xa8,0xeb, 0x59,0x23,0x70,0xad, 0x28,0xe2,0x9d,0x56, 0xe9,0x16,0xc4,0x65,
  0x62,0x93,0x54,0x60, 0xb0,0x39,0xb9,0x78, 0x1c,0x8d,0x68,0x36, 0x2c,0xeb,0x35,0x1a,
  0xdd,0x90,0xca,0x1c, 0x65,0x0c,0x16,0x09, 0x26,0x1a,0x30,0xdd, 0x09,0xa6,0x98,0xc4,
  0xc4,0xba,0x69,0x38, 0x21,0x87,0x14,0xc7, 0xab,0x48,0xbb,0x08, 0x86,0x73,0x40,0x6d,
  0x89,0x50,0x0d,0x6e, 0x21,0x48,0x92,0x28, 0x46,0xc8,0xab,0x14, 0x60,0xc8,0x4c,0xa2,
  0xee,0xef,0x5b,0x91, 0x61,0xae,0xf3,0x5a, 0x42,0x8d,0x0c,0x71, 0x29,0xaa,0x8d,0x46,
  0x19,0x26,0xb9,0x99, 0x3a,0x94,0xc0,0xa7, 0x1d,0x34,0x69,0x1c, 0x0b,0x23,0xb8,0xe3,
  0xce,0xcd,0x23,0xb5, 0x60,0xe0,0x4c,0xa2, 0xc9,0x92,0x73,0x82, 0xcc,0xb8,0x99,0x64,
  0x9d,0xa5,0xc9,0x17, 0x62,0xc6,0xe5,0x26, 0xaa,0xd3,0xb8,0x59, 0x97,0xbf,0x8c,0xed,
  // 0x1100:
  0x61,0xb9,0x83,0xb8, 0xcd,0x0b,0x3b,0xcc, 0xc9,0x93,0x13,0x2b, 0xc6,0xf8,0xd7,0x55,
  0x16,0x80,0xe1,0x42, 0xd0,0xb2,0x14,0xed, 0xdf,0x83,0x49,0x2d, 0xd8,0x96,0x51,0xd8,
  0x67,0x33,0x79,0x26, 0xd5,0x6b,0xc0,0xd8, 0xc2,0x32,0xff,0x9a, 0x7d,0x9f,0xb1,0xf5,
  0x7c,0x0e,0x07,0xef, 0xff,0x21,0x80,0xf0, 0x48,0xb6,0x58,0x28, 0x59,0xea,0xb1,0x86,
  0x4c,0x8c,0xb2,0x1e, 0x52,0xa4,0xf4,0x3d, 0x0d,0x6a,0x1b,0x21, 0xfb,0xd1,0xff,0xc1,
  0xd0,0xc1,0xf7,0x6d, 0xc9,0x23,0xde,0x90, 0xbe,0xd9,0x63,0xeb, 0xfe,0x27,0xd8,0xd6,
  0xff,0x25,0x3e,0xb8, 0xe8,0x6e,0x0f,0x88, 0x94,0xa5,0x91,0x67, 0xb7,0xd3,0x03,0xf3,
  0x93,0x40,0xe0,0x01, 0x58,0x0c,0x03,0x01, 0xcd,0x99,0x7d,0x5f, 0x59,0xf1,0x63,0xcd,
  0x50,0x01,0xe1,0x55, 0x00,0x90,0x11,0xdf, 0x72,0x17,0xee,0x99, 0x22,0x7c,0xb9,0x50,
  0x9e,0xe0,0xa9,0x09, 0xeb,0x3e,0x6d,0x0c, 0x7c,0x98,0x9e,0x91, 0x5c,0x8f,0x6c,0xc6,
  0x82,0xf1,0x2e,0x64, 0x30,0xe8,0x33,0x5b, 0x39,0x9d,0x73,0x39, 0xa8,0xce,0xaf,0x43,
  0x56,0x8d,0xd9,0xd5, 0x91,0x65,0xc0,0xeb, 0x5a,0x79,0xe6,0x48, 0xd8,0xd8,0xce,0xb5,
  0xcf,0x06,0xf0,0x39, 0xd6,0x86,0x86,0x3d, 0xda,0xef,0xbe,0xff, 0x00,0xde,0x4e,0xd9,
  0x9c,0x60,0xbc,0x75, 0x2d,0x3c,0x67,0x39, 0x38,0x71,0xcc,0xe0, 0x3f,0xe3,0x39,0x1c,
  0x58,0x52,0x6f,0x78, 0x73,0x2c,0x85,0xde, 0x39,0xa1,0xd0,0xe5, 0xbe,0x07,0xde,0xf3,
  0x81,0xbe,0x00,0x38, 0xae,0x1b,0x6d,0x43, 0x9e,0xeb,0x35,0x8b, 0x1d,0xd3,0xca,0x38,
  // 0x1200:
  0xbc,0xc7,0x2d,0x5d, 0x03,0x50,0x0e,0x49, 0xd4,0xad,0x5b,0xad, 0x6b,0x24,0x6a,0xb3,
  0xba,0x49,0x04,0xb0, 0x2e,0xb9,0xe2,0xe4, 0x09,0xcc,0xc5,0xe5, 0x35,0x03,0x74,0xeb,
  0x38,0x1c,0x62,0xae, 0x24,0xcc,0xdd,0x90, 0x16,0x62,0x06,0xa5, 0xa2,0xb1,0x80,0x90,
  0x92,0x44,0x43,0x86, 0x63,0x34,0x5d,0x52, 0xb8,0xcf,0x00,0x14, 0xca,0x0f,0x4f,0x8a,
  0xf6,0xdf,0xed,0x62, 0x4f,0xa5,0x3f,0x8b, 0xf6,0x74,0x8f,0xc6, 0x29,0x4b,0x23,0xb4,
  0x49,0x4a,0x14,0x3b, 0xca,0x91,0xe6,0x2c, 0x79,0x0b,0x13,0xd4, 0x35,0x94,0x54,0x51,
  0x49,0x19,0x53,0xcc, 0x6f,0x11,0x35,0x53, 0xd5,0x47,0xae,0xcb, 0x40,0xf5,0x14,0x48,
  0x70,0x14,0x43,0x25, 0x91,0x5e,0x20,0x28, 0x98,0x92,0x08,0x1d, 0x68,0x83,0x88,0x82,
  0xb5,0x04,0x16,0xe2, 0x86,0xd6,0x0f,0x16, 0x3b,0x59,0x3c,0x05, 0x23,0x12,0xdf,0x85,
  0x9b,0xb2,0x23,0xed, 0xc5,0x85,0x1b,0xc9, 0xe5,0x29,0x65,0x2a, 0xe7,0x8c,0x8e,0x2c,
  0x2c,0xdf,0x8f,0xa3, 0xef,0xfa,0x39,0xca, 0x8e,0xb3,0x0d,0x4b, 0x15,0x1d,0x80,0x40,
  0xd3,0x1c,0xf9,0x74, 0x83,0xcf,0x17,0x4c, 0x47,0xa0,0x02,0x22, 0x27,0x00,0x46,0x13,
  0x9b,0x22,0x1a,0x96, 0x41,0x96,0x23,0x54, 0x65,0x89,0x24,0xb6, 0x3d,0x25,0x36,0x18,
  0x68,0x33,0x21,0xca, 0x9e,0xa1,0x2a,0x32, 0x0f,0xa3,0x08,0x23, 0xfa,0xa0,0x41,0x39,
  0x33,0x6e,0xf7,0xcf, 0x1d,0x78,0xe0,0x78, 0xbe,0x5c,0xe4,0xe7, 0x6c,0x8e,0x26,0xf0,
  0xa3,0xb2,0xdd,0x69, 0x03,0x25,0x89,0x24, 0xd3,0x45,0x31,0xcf, 0x93,0x7e,0x2c,0x27,
  // 0x1300:
  0x8e,0xfc,0xfd,0x80, 0xeb,0xed,0x6f,0x50, 0xd8,0x7b,0x98,0x01, 0xca,0x02,0x3d,0xde,
  0x54,0xa2,0x9b,0x24, 0x08,0x30,0xeb,0x1d, 0x00,0x71,0x7e,0x4d, 0x4c,0x83,0x47,0x43,
  0x5a,0x1f,0xf8,0xbb, 0x92,0x29,0xc2,0x84, 0x83,0xa9,0xf0,0x61, 0x50,               
  // 0x132d = 4909 Bytes
};

//
///////////////////////////////////////////////////////////////////////////////

static const u8 ctcode_brres_strings_bz2[] =
{
  0x00,0x00,0x01,0x40, 0x42,0x5a,0x68,0x31, 0x31,0x41,0x59,0x26, 0x53,0x59,0x0f,0x63,
  0x56,0x7c,0x00,0x00, 0x09,0xff,0x80,0x70, 0x21,0x08,0x00,0x00, 0x60,0x7f,0x60,0x3e,
  0x81,0xd4,0x80,0xa2, 0x00,0x5e,0x40,0x20, 0x00,0x75,0x09,0x54, 0xf4,0x40,0x00,0x0d,
  0x00,0xd3,0x20,0x02, 0xa8,0xa8,0x3d,0x40, 0x00,0x68,0x06,0x8d, 0x34,0xda,0x9b,0x51,
  0x9e,0x6c,0xc6,0xc5, 0x31,0x1f,0x66,0x02, 0x92,0xa2,0x2f,0xf5, 0x43,0xe3,0x81,0xcc,
  0xa8,0xba,0xc1,0xd0, 0x23,0x06,0x08,0x0c, 0x04,0x3c,0x04,0xc1, 0x43,0x90,0x6f,0x5a,
  0xf2,0xea,0x55,0xb3, 0xdc,0xf2,0x19,0x96, 0xf1,0x6b,0xc8,0xe9, 0xb0,0x06,0xcb,0xc5,
  0xb8,0x0a,0xd7,0x0d, 0xf3,0x30,0xb8,0x5b, 0x4a,0xbe,0x26,0x53, 0x9b,0xf8,0xbb,0x92,
  0x29,0xc2,0x84,0x80, 0x7b,0x1a,0xb3,0xe0,                                          
  // 0x88 = 136 Bytes
};

//
///////////////////////////////////////////////////////////////////////////////
///////////////			ctcode_data_tab[]		///////////////
///////////////////////////////////////////////////////////////////////////////

#undef DEF_DATA
#undef DEF_BZIP
//#undef DEF_PATCH

#define DEF_DATA(a) { #a, 0, a, sizeof(a) },
#define DEF_BZIP(a) { #a, 1, a##_bz2, sizeof(a##_bz2) },
//#define DEF_PATCH(m,a,b) { #a, m, b, sizeof(b), a },

const data_tab_t ctcode_data_tab[] =
{
	DEF_DATA(ctcode_header)
	DEF_DATA(cup1_header)
	DEF_DATA(crs1_header)

	DEF_BZIP(ctcode_boot_code_pal)
	DEF_BZIP(ctcode_boot_code_usa)
	DEF_BZIP(ctcode_boot_code_jap)

	DEF_BZIP(ctcode_boot_data_pal)
	DEF_BZIP(ctcode_boot_data_usa)
	DEF_BZIP(ctcode_boot_data_jap)

	DEF_BZIP(ctcode_bad1code_pal)
	DEF_BZIP(ctcode_bad1code_usa)
	DEF_BZIP(ctcode_bad1code_jap)

	DEF_BZIP(ctcode_mod1_pal)
	DEF_BZIP(ctcode_mod1_usa)
	DEF_BZIP(ctcode_mod1_jap)

	DEF_BZIP(ctcode_mod2_pal)
	DEF_BZIP(ctcode_mod2_usa)
	DEF_BZIP(ctcode_mod2_jap)

	DEF_BZIP(ctcode_ovr1_pal)
	DEF_BZIP(ctcode_ovr1_usa)
	DEF_BZIP(ctcode_ovr1_jap)

	DEF_BZIP(ctcode_image)

	DEF_BZIP(ctcode_brres_base)
	DEF_BZIP(ctcode_brres_strings)

	{0,0,0}
};

#undef DEF_DATA
#undef DEF_BZIP
//#undef DEF_PATCH

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    END				///////////////
///////////////////////////////////////////////////////////////////////////////

