
/***************************************************************************
 *                         _______ _______ _______                         *
 *                        |  ___  |____   |  ___  |                        *
 *                        | |   |_|    / /| |   |_|                        *
 *                        | |_____    / / | |_____                         *
 *                        |_____  |  / /  |_____  |                        *
 *                         _    | | / /    _    | |                        *
 *                        | |___| |/ /____| |___| |                        *
 *                        |_______|_______|_______|                        *
 *                                                                         *
 *                             Wiimms SZS Tools                            *
 *                           http://szs.wiimm.de/                          *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This file is part of the SZS project.                                 *
 *   Visit http://szs.wiimm.de/ for project details and sources.           *
 *                                                                         *
 *   Copyright (c) 2011-2016 by Dirk Clemens <wiimm@wiimm.de>              *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   See file gpl-2.0.txt or http://www.gnu.org/licenses/gpl-2.0.txt       *
 *                                                                         *
 ***************************************************************************/

#ifndef SZS_LIB_NUMERIC_H
#define SZS_LIB_NUMERIC_H 1

#define _GNU_SOURCE 1

#include "dclib-numeric.h"

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  defines			///////////////
///////////////////////////////////////////////////////////////////////////////

#define EXPAND8(a) ( (a) + 7 & ~(u32)7 )

///////////////////////////////////////////////////////////////////////////////

#define MAX_LOOP_DEPTH		10 // max loop depth
#define MAX_FUNCTION_DEPTH	10 // max macro depth
#define MAX_SOURCE_DEPTH	50 // max open source files/macros

///////////////////////////////////////////////////////////////////////////////

struct Var_t;
struct FuncParam_t;
struct ScanMacro_t;
struct ScanFile_t;
struct ScanInfo_t;

//
///////////////////////////////////////////////////////////////////////////////
///////////////			scan options			///////////////
///////////////////////////////////////////////////////////////////////////////

int ScanOptAlignU8	( ccp arg );
int ScanOptAlignPACK	( ccp arg );
int ScanOptAlignBRRES	( ccp arg );
int ScanOptAlignBREFF	( ccp arg );
int ScanOptAlignBREFT	( ccp arg );
int ScanOptAlign	( ccp arg );
int ScanOptMaxFileSize	( ccp arg );
int ScanOptNMipmaps	( ccp arg );
int ScanOptMipmapSize	( ccp arg );
int ScanOptPtDir	( ccp arg );
int ScanOptRecurse	( ccp arg );

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  scan num/range		///////////////
///////////////////////////////////////////////////////////////////////////////

double strtod_comma ( const char *nptr, char **endptr );

///////////////////////////////////////////////////////////////////////////////

char * ScanNumU32
(
    ccp			arg,		// text to scan (valid pointer)
    u32			* p_stat,	// NULL or return status:
					//   number of scanned numbers (0..1)
    u32			* p_num,	// return value (valid pointer)
    u32			min,		// min allowed value (fix num)
    u32			max		// max allowed value (fix num)
);

char * ScanRangeU32
(
    ccp			arg,		// text to scan (valid pointer)
    u32			* p_stat,	// NULL or return status:
					//   number of scanned numbers (0..2)
    u32			* p_n1,		// first return value (valid pointer)
    u32			* p_n2,		// second return value (valid pointer)
    u32			min,		// min allowed value (fix num)
    u32			max		// max allowed value (fix num)
);

unsigned long int strtoul16
(
    const u16		*  ptr16,
    u16			** end16,
    int			base
);

u16 * skip_ctrl16 ( const u16 * ptr );

u16 * ScanNum16U32
(
    const u16		* arg,		// text to scan (valid pointer)
    u32			* p_stat,	// NULL or return status:
					//   number of scanned numbers (0..1)
    u32			* p_num,	// return value (valid pointer)
    u32			min,		// min allowed value (fix num)
    u32			max		// max allowed value (fix num)
);

u16 * ScanRange16U32
(
    const u16		* arg,		// text to scan (valid pointer)
    u32			* p_stat,	// NULL or return status:
					//   number of scanned numbers (0..2)
    u32			* p_n1,		// first return value (valid pointer)
    u32			* p_n2,		// second return value (valid pointer)
    u32			min,		// min allowed value (fix num)
    u32			max		// max allowed value (fix num)
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			vector calculations		///////////////
///////////////////////////////////////////////////////////////////////////////

double LengthF
(
    const float *a,		// A vector (x,y,z)
    const float *b		// B vector (x,y,z)
);

double LengthD
(
    const double *a,		// A vector (x,y,z)
    const double *b		// B vector (x,y,z)
);

double Length2D
(
    const double *a,		// A vector (x,y,z)
    const double *b		// B vector (x,y,z)
);

double AngleVector
(
    const float	* prev,		// vector (x,y,z) of previous point
    const float	* point,	// vector (x,y,z) of current point
    const float	* next,		// vector (x,y,z) of next point
    bool	return_degree	// true: return in normed degree; false: radiants
);

void WidthAndDirection
(
    double	* r_width,	// if not NULL: return width here
    double	* r_direction,	// if not NULL: return direction here
    bool	return_degree,	// true: return in normed degree; false: radiants
    const float	* a,		// A point (x,y)
    const float	* b		// B point (x,y)
);

double CalcDirection2F
(
    const float	* a,		// A point (x,y)
    const float	* b		// B point (x,y)
);

float3 CalcDirection3F
(
    const float * a,		// A point (x,y,z)
    const float * b		// B point (x,y,z)
);

double3 CalcDirection3D
(
    const double3 * a,		// A point (x,y,z)
    const double3 * b		// B point (x,y,z)
);

void Scale
(
    const double3	*origin,	// origin of rotation, NULL = v(0,0,0)
    const double3	*scale,		// NULL or scaling vector
    double3		*pt,		// pointer to first point to rotate
    uint		pt_delta,	// delta in bytes to next point
    uint		n_pt		// number of points to rotate
);

void Rotate
(
    const double3	*origin,	// origin of rotation, NULL = v(0,0,0)
    const double3	*rot,		// rotation vector
    double3		*pt,		// pointer to first point to rotate
    uint		pt_delta,	// delta in bytes to next point
    uint		n_pt		// number of points to rotate
);

void Translate
(
    const double3	*trans,		// NULL or translation vector
    double3		*pt,		// pointer to first point to rotate
    uint		pt_delta,	// delta in bytes to next point
    uint		n_pt		// number of points to rotate
);

void AxisRotate
(
    const double3	*a1,		// first axis point
    const double3	*a2,		// second axis point
    double		degree,		// degree of rotation
    double3		*pt,		// pointer to first point to rotate
    uint		pt_delta,	// delta in bytes to next point
    uint		n_pt		// number of points to rotate
);

double Degree // output: -180.00 .. +180.00
(
    double	rad		// radians
);

float NormDegreeF // output: -180.00 .. +180.00
(
    float	deg
);

void MinMax3
(
    double3		*min,		// valid pointer to initialized min value
    double3		*max,		// valid pointer to initialized max value
    const double3	*val,		// valid pointer to source list
    int			n_val		// if >0: number of elements in 'val'
);

void UnitZero3 ( double3 *dest, const double3 *src );

double CalcNormals
(
    // return length of p1..p2

    double3		*res_norm,	// return 3 normals here
    const double3	*p1,		// first point
    const double3	*p2,		// second point
    const double3	*dir,		// NULL or helper point for normal[1]
    double		r		// radius == length of norm[0]..norm[2]
);

bool OverlapCubeTri
(
    double3		*cube_mid,	// middle of the cube
    double		cube_width,	// width of the cube
    double3		*pt1,		// first point of triangle
    double3		*pt2,		// second point of triangle
    double3		*pt3		// third point of triangle
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  octahedron_t			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct octahedron_t
{
    union
    {
	double3 p[6];
	struct
	{
	    double3 x1;	 // left point
	    double3 x2;	 // right point
	    double3 y1;  // bottom point
	    double3 y2;  // top point
	    double3 z1;  // far point
	    double3 z2;  // near point
	};
    };

} octahedron_t;

//-----------------------------------------------------------------------------

void CreateRegularOctahedron
(
    octahedron_t	*oct,		// valid pointer, initialize
    double		radius,		// radius of octahedron
    const double3	*rot,		// not NULL: rotate octahedron
    const double3	*shift		// not NULL: shift octahedron
);

//-----------------------------------------------------------------------------

void CreateRegularOctahedronF
(
    octahedron_t	*oct,		// valid pointer, initialize
    double		radius,		// radius of octahedron
    const float		*rot,		// not NULL: 3D vector, rotate octahedron
    const float		*shift		// not NULL: 3D vector, shift octahedron
);

//-----------------------------------------------------------------------------

void CreateOctahedron
(
    octahedron_t	*oct,		// valid pointer, initialize
    const double3	*radius,	// radius of octahedron
    const double3	*rot,		// not NULL: rotate octahedron
    const double3	*shift		// not NULL: shift octahedron
);

//-----------------------------------------------------------------------------

void ScaleOctahedron
(
    octahedron_t	*oct,		// valid octahedron data
    const double3	*origin,	// origin of rotation, NULL = v(0,0,0)
    const double3	*scale		// not NULL: scale octahedron
);

//-----------------------------------------------------------------------------

void RotateOctahedron
(
    octahedron_t	*oct,		// valid octahedron data
    const double3	*origin,	// origin of rotation, NULL = v(0,0,0)
    const double3	*rot		// not NULL: rotate octahedron
);

//-----------------------------------------------------------------------------

void ShiftOctahedron
(
    octahedron_t	*oct,		// valid octahedron data
    const double3	*shift		// not NULL: shift octahedron
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			  cuboid_t			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct cuboid_t
{
    union
    {
	double3 p[8];
	struct
	{
	    double3 p000;
	    double3 p001;
	    double3 p010;
	    double3 p011;
	    double3 p100;
	    double3 p101;
	    double3 p110;
	    double3 p111;
	};
    };

} cuboid_t;

//-----------------------------------------------------------------------------

void CreateCube
(
    cuboid_t		*cub,		// valid pointer, initialize
    double		width,		// width of cube
    const double3	*rot,		// not NULL: rotate cube
    const double3	*shift		// not NULL: shift cube
);

//-----------------------------------------------------------------------------

void CreateCubeF
(
    cuboid_t		*cub,		// valid pointer, initialize
    double		width,		// width of cube
    const float		*rot,		// not NULL: 3D vector, rotate cube
    const float		*shift		// not NULL: 3D vector, shift cube
);

//-----------------------------------------------------------------------------

void CreateCuboid
(
    cuboid_t		*cub,		// valid pointer, initialize
    const double3	*size,		// base size of cuboid
    const double3	*rot,		// not NULL: rotate cuboid
    const double3	*shift		// not NULL: shift cuboid
);

//-----------------------------------------------------------------------------

void CreateCuboidF
(
    cuboid_t		*cub,		// valid pointer, initialize
    const float		*size,		// base size of cuboid
    const float		*rot,		// not NULL: rotate cuboid
    const float		*shift		// not NULL: shift cuboid
);

//-----------------------------------------------------------------------------

void ScaleCuboid
(
    cuboid_t		*cub,		// valid cuboid data
    const double3	*origin,	// origin of rotation, NULL = v(0,0,0)
    const double3	*scale		// not NULL: scale cuboid
);

//-----------------------------------------------------------------------------

void RotateCuboid
(
    cuboid_t		*cub,		// valid cuboid data
    const double3	*origin,	// origin of rotation, NULL = v(0,0,0)
    const double3	*rot		// not NULL: rotate cuboid
);

//-----------------------------------------------------------------------------

void ShiftCuboid
(
    cuboid_t		*cub,		// valid cuboid data
    const double3	*shift		// not NULL: shift cuboid
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			index lists			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef struct IndexList_t
{
    ccp		* list;		// list of names
    uint	used;		// used elements of 'list'
    uint	size;		// alloced elements of 'list'

    ccp		prefix;		// prefix for automatic named variables
    uint	next_index;	// next index for autmatic named variables

} IndexList_t;

//-----------------------------------------------------------------------------

void InitializeIL ( IndexList_t * il, ccp prefix );
void ResetIL ( IndexList_t * il );

void FillIL
(
    IndexList_t		* il,		// valid index list
    uint		min_fill	// fill minimum # elements
);

void DefineIL
(
    IndexList_t		* il,		// valid index list
    uint		index,		// index of object
    ccp			name,		// name of object
    bool		move_name	// true: 'name' is alloced -> move it
);

bool RemoveIL
(
    IndexList_t		* il,		// valid index list
    uint		index		// index of object
);

ccp GetNameIL
(
    IndexList_t		* il,		// valid index list
    uint		index		// index of object
);

void PrintNameIL
(
    FILE		* f,		// output files
    IndexList_t		* il,		// valid index list
    uint		index,		// index of object
    uint		align,		// align end of name to character pos
    uint		fw,		// field width of name
    uint		indent		// indention of continuation line
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			parser functions		///////////////
///////////////////////////////////////////////////////////////////////////////

#define MAX_FUNC_PARAM 1000

typedef enumError (*call_func_t)
(
    struct Var_t	* res,		// store result here
    struct Var_t	* param,	// parameters, modified
    uint		n_param,	// number of parameters
    struct ScanInfo_t	* si,		// not NULL: use for error messages
    const struct FuncParam_t
			* fpar		// pointer to user parameter
);

typedef struct FuncParam_t
{
    bool		need_name;	// true: needs a name as first param
    s16			min_param;	// minimum parameter
    s16			max_param;	// maximum parameter
    u16			fform_id;	// function class, based on file_format_t
    union
    {
	s32		user_id;	// user defined id
	struct ScanMacro_t
			* macro;	// NULL or related macro to call
    };
    call_func_t		func;		// function to call

} FuncParam_t;

extern FuncParam_t FuncDollar;

//-----------------------------------------------------------------------------

typedef struct FuncTable_t
{
    //--- function def
    s32			min_param;	// minimum parameter,
					// <0: first param is a var name
    s32			max_param;	// maximum parameter
    ccp			name;		// name of function, NULL: info only
    call_func_t		func;		// function to call, NULL: end of list
    int			user_id;	// user defined id

    //--- function info

    ccp			result;		// function result info
    ccp			syntax;		// syntax of function, NULL: no info
    ccp			info;		// description

} FuncTable_t;

//-----------------------------------------------------------------------------

void DefineDefaultParserFunc();

void DefineParserFuncTab
(
    const FuncTable_t	*ftab,		// pointer to a definittion table
					// terminated with member 'func == 0'
    u16			fform_id	// function class, based on file_format_t
);

FuncParam_t * DefineParserFunc
(
    ccp			name,		// name of function
    s16			min_param,	// minimum parameter
    s16			max_param,	// maximum parameter
    u16			fform_id,	// function class, based on file_format_t
    call_func_t		func,		// function to call
    int			user_id		// user defined id
);

const FuncParam_t * GetParserFunc
(
    ccp			name,		// name to search
    struct ScanInfo_t	*si,		// not NULL: print warning if not found
    FuncParam_t		*temp_par	// not NULL: search user defined functions,
					// and use this to create the function result
);

///////////////////////////////////////////////////////////////////////////////

typedef struct FuncInfo_t
{
    ccp			result;		// function result info
    ccp			info;		// description

} FuncInfo_t;

struct search_filter_t;

//-----------------------------------------------------------------------------

FuncInfo_t * DefineParserFuncInfo
(
    ccp			result,		// function result info
    ccp			syntax,		// syntax of function
    ccp			info		// description
);

void DumpParserFuncInfo
(
    FILE		* f,		// destination file
    uint		indent,		// indention
    bool		print_header,	// true: print table header
    int			long_count,	// >0: print description too
    struct search_filter_t * filter	// NULL or filter data
);

//-----------------------------------------------------------------------------

struct search_filter_t * SetupParserFuncFilter();
void ParserFuncFilter ( struct search_filter_t *filter_data, ccp arg, int mode );

void ExportParserFuncMakedoc ( FILE * f, ccp index );

//
///////////////////////////////////////////////////////////////////////////////
///////////////			variables & maps		///////////////
///////////////////////////////////////////////////////////////////////////////

#define VARNAME_SIZE	100
#define STACK_SIZE	100

typedef enum VarMode_t
{
    VAR_UNSET,
    VAR_INT,
    VAR_DOUBLE,
    VAR_VECTOR

} VarMode_t;

typedef enum VarModeHelpers_t
{
    VAR_X = VAR_VECTOR+1,
    VAR_Y,
    VAR_Z

} VarModeHelpers_t;

//-----------------------------------------------------------------------------
// [[Var_t]]

typedef struct Var_t
{
    ccp		name;	// name of variable, alloced
    VarMode_t	mode;	// mode of variable

    union
    {
	int	i;	// int value
	double	d;	// double value
	double  v[3];	// double vector
	struct
	{
	    double x;	// double components
	    double y;
	    double z;
	};
	double3 d3;	// double3 value for references

	// for internal usage
	FuncParam_t func_param;
	FuncInfo_t  func_info;
	struct ScanMacro_t * macro;
    };

} Var_t;

//-----------------------------------------------------------------------------

typedef struct VarMap_t
{
    Var_t	* list;			// pointer to the item list
    uint	used;			// number of used elements of 'list'
    uint	size;			// number of allocated  elements of 'list'

} VarMap_t;

void DefineParserVars ( VarMap_t * vm );
const VarMap_t * SetupParserVars();

enumError DumpSymbols ( const VarMap_t * vm );
enumError ListParserFunctions();
enumError ParserCalc ( const VarMap_t * predef );

//
///////////////////////////////////////////////////////////////////////////////
///////////////			scan text			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef enum TextCommand_t
{
    TCMD_NONE		= 0,	// definitiv NULL

    TCMD_ECHO,
    TCMD_WARN,

    TCMD_DUMP_VAR,
    TCMD_DUMP_PRIVATE,
    TCMD_DUMP_LOCAL,
    TCMD_DUMP_GLOBAL,
    TCMD_DUMP_CONST,
    TCMD_DUMP_PREDEF,

    TCMD_REVISION,
    TCMD_SZS_MODIFIER,
    TCMD_SETUP_RANDOM,

    TCMD_PDEF,		//+++ 'TCMD_LDEF* - TCMD_PDEF*' is constant for all '*'
    TCMD_PDEF_I,	//
    TCMD_PDEF_F,	//
    TCMD_PDEF_V,	//
    TCMD_PDEF_X,	//	//+++ _X _Y and _Z are well sorted
    TCMD_PDEF_Y,	//	//
    TCMD_PDEF_Z,	//	//---
    TCMD_PDEF_ENUM,	//
    TCMD_PDEF_SHIFT,	//---

    TCMD_LDEF,		//+++ 'TCMD_PDEF* - TCMD_LDEF*' is constant for all '*'
    TCMD_LDEF_I,	//
    TCMD_LDEF_F,	//
    TCMD_LDEF_V,	//
    TCMD_LDEF_X,	//	//+++ _X _Y and _Z are well sorted
    TCMD_LDEF_Y,	//	//
    TCMD_LDEF_Z,	//	//---
    TCMD_LDEF_ENUM,	//
    TCMD_LDEF_SHIFT,	//---

    TCMD_GDEF,		//+++ 'TCMD_GDEF* - TCMD_LDEF*' is constant for all '*'
    TCMD_GDEF_I,	//
    TCMD_GDEF_F,	//
    TCMD_GDEF_V,	//
    TCMD_GDEF_X,	//	//+++ _X _Y and _Z are well sorted
    TCMD_GDEF_Y,	//	//
    TCMD_GDEF_Z,	//	//---
    TCMD_GDEF_ENUM,	//
    TCMD_GDEF_SHIFT,	//---

    TCMD_IF,		//+++ first IF related command
    TCMD_ELIF,		//
    TCMD_ELSE,		//
    TCMD_ENDIF,		//--- last IF related command

    TCMD_DOIF,

    TCMD_LOOP,		//+++ LOOP commands, same order as TCMD_END* & loop_name[]
    TCMD_REPEAT,	//
    TCMD_FOR,		//
    TCMD_FOREACH,	//
    TCMD_WHILE,		//---

    TCMD_ENDLOOP,	//+++ END-LOOP commands, same order as TCMD_LOOP...
    TCMD_ENDREPEAT,	//    & loop_name[]
    TCMD_ENDFOR,	//
    TCMD_ENDFOREACH,	//
    TCMD_ENDWHILE,	//---

    TCMD_BREAK,		// terminate a loop
    TCMD_CONTINUE,	// continue a loop

    TCMD_INCLUDE,
    TCMD_RETURN,
    TCMD_EXIT,
    TCMD_ASSERT,

    TCMD_MACRO,		//+++ same order as TCMD_END*
    TCMD_FUNCTION,	//---
    TCMD_ENDMACRO,	//+++ same order as TCMD_* without END
    TCMD_ENDFUNCTION,	//---

    TCMD_CALL,
    TCMD_PARAMETERS,

    TCMD__N

} TextCommand_t;

//-----------------------------------------------------------------------------

typedef struct ScanLoop_t
{
    int			tcmd_end;	// closing TCMD_*
    ccp			ptr;		// file position of loop start
    uint		line;		// line number on loop begin
    int			if_level;	// number of active IF..ENDIF on loop begin

    int			count_val;	// internal loop counter 0..
    int			end_val;	// end value for @REPEAT + @FOR
    int			step_val;	// step_val for @FOR
    ccp			varname;	// variable name for @FOR + @FOREACH, alloced
    ccp			expr_ptr;	// expression pointer
    uint		expr_line;	// expression line

} ScanLoop_t;

//-----------------------------------------------------------------------------

typedef struct ScanMacro_t
{
    ccp			data;		// begin of text
    uint		data_size;	// size of data
    bool		data_alloced;	// true if 'data' alloced
    uint		line0;		// line number of first line
    ccp			src_name;	// name of source file, alloced
    bool		is_function;	// true: defined as function
    Var_t		* map;		// pointer to map entry

} ScanMacro_t;

//-----------------------------------------------------------------------------

typedef struct ScanFile_t // includes and macros
{
    //--- scanning infos

    struct ScanInfo_t	* si;		// base scan info
    ScanMacro_t		* sm;		// not NULL: pointer to macro

    ccp			data;		// begin of text
    bool		data_alloced;	// true if 'data' alloced
    ccp			ptr;		// current text position
    ccp			end;		// end of text
    ccp			prev_ptr;	// 'ptr' before scanning an object
    uint		line;		// current line number
    uint		line_err;	// number of line errors
    ccp			name;		// name or file for warnings
    bool		name_alloced;	// true if 'name' alloced
    int			if_level;	// number of active IF..ENDIF
    struct ScanFile_t	* next;		// next file for RETURN

    //--- loop control

    int			loop_level;	// number of active loops
    ScanLoop_t		loop[MAX_LOOP_DEPTH];	// loop data

    //--- some options

    int			revision;	// current revision number
    int			szs_modifier;	// >0: szs_modifier mode enabled
    int			disable_comma;	// >0: disable comma in floats
    int			expr_depth;	// level of expression depth

    //--- variables

    VarMap_t		pvar;		// private variables
    const ScanMacro_t	* active_macro;	// active macro or function

    //--- parameter stack

    Var_t		param[STACK_SIZE]; // parameter stack
    int			param_used;	// number of used parameters

} ScanFile_t;

extern ScanFile_t empty_scan_file;

//-----------------------------------------------------------------------------
// [[ScanInfo_t]]

typedef struct ScanInfo_t
{
    //--- init data

    ccp			init_data;	// begin of text
    uint		init_data_size;	// end of text
    ccp			init_name;	// name or file for warnings
    int			init_revision;	// revision number at start

    //--- current file

    uint		n_files;	// number of open files and macros
    ScanFile_t		* cur_file;	// pointer to current file, never NULL
    VarMap_t		macro;		// macro map

    //--- scanning infos

    uint		total_err;	// total number of errors
    int			no_warn;	// >0: suppress warnings
    int			no_calc;	// >0: suppress calculations and function calls
    int			fast_scan;	// >0: ignore @commands
    int			force_hex;	// >0: force hex for non prefixed integers
 #if HAVE_PRINT
    int			debug;		// >0: print debugging infos
 #endif

    //--- variables

    VarMap_t		* param;	// NULL or parameters, not cleared
    VarMap_t		lvar;		// local variables
    VarMap_t		gvar;		// global variables
    const VarMap_t	* predef;	// NULL or predefined variables
    Var_t		last_result;	// last macro/function result
    int			func_status;	// second status of some parser functions

    //--- user info

    struct kcl_t	* kcl;		// for kcl scanning: KCL data
    struct kmp_t	* kmp;		// for kmp scanning: KMP data

} ScanInfo_t;

//-----------------------------------------------------------------------------

typedef enum ScanParamMode
{
    SPM_NONE,		// ignore result

    SPM_BOOL,		// result has type bool

    SPM_U8,		// result has type u8
    SPM_S8,		// result has type s8
    SPM_U16,		// result has type u16
    SPM_S16,		// result has type s16
    SPM_U32,		// result has type u32
    SPM_S32,		// result has type s32

    SPM_INT,		// result has type int
    SPM_UINT,		// result has type uint
    SPM_FLOAT,		// result has type float
    SPM_DOUBLE,		// result has type double

    SPM_DOUBLE_X,	// result has type double, use X on vector
    SPM_DOUBLE_Y,	// result has type double, use Y on vector
    SPM_DOUBLE_Z,	// result has type double, use Z on vector

    SPM_STRING,		// result has type char*
    SPM_VAR,		// result has type Var_t

}  ScanParamMode;

//-----------------------------------------------------------------------------

typedef struct ScanParam_t
{
    ccp			name;		// name of the parameter
    ScanParamMode	mode;		// type of result
    void		*result;	// pointer to result
    uint		min;		// >1: scan at least 'min' values
    uint		max;		// >min: scan maxixmal max values
    int			*n_result;	// not NULL: store number of read values

} ScanParam_t;

//
///////////////////////////////////////////////////////////////////////////////
///////////////			conversion helpers		///////////////
///////////////////////////////////////////////////////////////////////////////

bool	GetBoolV    ( const Var_t * var );
int	GetIntV	    ( const Var_t * var );
double	GetDoubleV  ( const Var_t * var );
double	GetXDoubleV ( const Var_t * var );
double	GetYDoubleV ( const Var_t * var );
double	GetZDoubleV ( const Var_t * var );
double3	GetVectorV  ( const Var_t * var );
float3	GetVectorFV ( const Var_t * var );

Var_t * ToBoolV    ( Var_t * var );
Var_t * ToIntV     ( Var_t * var );
Var_t * ToDoubleV  ( Var_t * var );
Var_t * ToXDoubleV ( Var_t * var );
Var_t * ToYDoubleV ( Var_t * var );
Var_t * ToZDoubleV ( Var_t * var );
Var_t * ToVectorV  ( Var_t * var );
Var_t * ToVectorXV ( Var_t * var );
Var_t * ToVectorYV ( Var_t * var );
Var_t * ToVectorZV ( Var_t * var );
Var_t * ToVarMode  ( Var_t * var, TextCommand_t mode );

void AssignIntV     ( Var_t * var, int val );
void AssignDoubleV  ( Var_t * var, double val );
void AssignXDoubleV ( Var_t * var, double val );
void AssignYDoubleV ( Var_t * var, double val );
void AssignZDoubleV ( Var_t * var, double val );
void AssignVectorV  ( Var_t * var, const double3 * val );
void AssignVectorV3 ( Var_t * var, double x, double y, double z );
void AssignVar      ( Var_t * dest, const Var_t * src );
void AssignVarMode  ( Var_t * dest, const Var_t * src, TextCommand_t mode );

//--- special assigning helpers

enumError ResultFloat
(
    Var_t		* dest,		// valid destination var
    const float		* val,		// pointer to N floats
    uint		n_val		// number of floats in 'n'
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////		variables & maps: interface		///////////////
///////////////////////////////////////////////////////////////////////////////

void InitializeVarMap
(
    VarMap_t		* vm		// data structure to initialize
);

//-----------------------------------------------------------------------------

void ResetVarMap
(
    VarMap_t		* vm		// data structure to resewt
);

//-----------------------------------------------------------------------------

void ClearVarMap
(
    VarMap_t		* vm		// data structure to resewt
);

//-----------------------------------------------------------------------------

void CopyVarMap
(
    VarMap_t		* dest,		// destination var map
    bool		init_dest,	// true: initialialize 'dest' first
    const VarMap_t	* src		// source var map
);

//-----------------------------------------------------------------------------

void MoveVarMap
(
    VarMap_t		* dest,		// destination var map
    bool		init_dest,	// true: initialialize 'dest' first
    VarMap_t		* src		// source var map -> will be resetted
);

//-----------------------------------------------------------------------------

int FindVarMapIndex
(
    const VarMap_t	* vm,		// valid variable map
    ccp			varname,	// variable name
    int			not_found_value	// return value if not found
);

//-----------------------------------------------------------------------------

bool RemoveVarMap
(
    // returns 'true' if var deleted

    VarMap_t		* vm,		// valid variable map
    ccp			varname	// variable name
);

//-----------------------------------------------------------------------------

Var_t * InsertVarMap
(
    VarMap_t		* vm,		// valid variable map
    ccp			varname,	// varname to insert
    bool		move_varname,	// true: move 'varname', false: strdup()
    ScanInfo_t		* si,		// not null: print "already defined" message
    bool		* found		// not null: store found status
);

//-----------------------------------------------------------------------------

bool DefineIntVar
(
    VarMap_t		* vm,		// valid variable map
    ccp			varname,	// name of variable
    int			value		// value to assign
);

//-----------------------------------------------------------------------------

bool DefineDoubleVar
(
    VarMap_t		* vm,		// valid variable map
    ccp			varname,	// name of variable
    double		value		// value to assign
);

//-----------------------------------------------------------------------------

const Var_t * FindConst
(
    ccp			varname	// variable name
);

//-----------------------------------------------------------------------------

const Var_t * FindVarMap
(
    const VarMap_t	* vm,		// NULL or variable map
    ccp			varname,	// variable name
    ScanInfo_t		* si		// not null: print "not found" message
);

//-----------------------------------------------------------------------------

void PrintValue
(
    FILE		* f,		// destination file
    const Var_t		*v		// variable to print
);

//-----------------------------------------------------------------------------

void DumpVarMap
(
    FILE		* f,		// destination file
    uint		indent,		// indention
    const VarMap_t	* vm,		// valid variable map
    bool		print_header	// print table header
);

//-----------------------------------------------------------------------------

extern VarMap_t const_map;

#define MINIMAP_DEFAULT_FLAGS 0x31f

extern bool  have_set_scale;
extern bool  have_set_rot;
extern Var_t set_scale, set_rot;
extern u32   set_flags;

extern uint	have_set_value; // x=bit-0, y=bit-1, z=bit-2
extern double3	set_value_min, set_value_max;

//-----------------------------------------------------------------------------
// transformation options, values and functions

#define N_TRANSFORM 10

extern uint	force_transform;		// bit 0: force tform even if not needed
						// bit 1: ignore it

extern MatrixD_t opt_transform;			// current transform values
extern MatrixD_t transform_list[N_TRANSFORM];	// list with stored tforms
extern uint	n_transform_list;		// number of elements in list
extern bool	transform_active;		// true: transformation active
extern int	disable_transformation;		// >0: transformation is temporary disabled

extern double	yposition;			// defined y position
extern bool	yposition_valid;		// 'yposition' is valid

extern bool	ascale_valid;			// true: --ascale defined
extern double	ascale;				// scaling factor
extern double3	ascale_dir;			// scaling direction (vector)

extern bool	arot_valid;			// true: --arot defined
extern double	arot_deg;			// angle of arot in degree
extern double3	arot1, arot2;			// points to define rotation axis

extern uint	mdl_switched_to_vertex;		// 0: default
						// 1: switch to 'VERTEX' needed
						// 2: MDL_MODE switched to 'VERTEX'
						// 3: '2' and message printed

//------------------------------

void ResetTransformation();			// reset all transformation steps
int  NextTransformation( bool for_close );	// enter next transformation step
void CloseTransformation();			// close transformation input
void DumpTransformationOpt();			// dump transformation options

double3 TransformD3( uint mode, double3 *val );	// transformate a singe vector
// mode:
//   0: use opt_transform and default
//   1: single steps through 'list', use base vectors
//   2: single steps through 'list', use norm vectors
//   3: single steps through 'list', use matrix

int ScanOptShift ( ccp arg );			// --shift
int ScanOptTranslate ( ccp arg );		// --translate
int ScanOptScale ( ccp arg );			// --scale
int ScanOptRotate ( ccp arg );			// --rot
int ScanOptXRotate ( uint xyz, ccp arg );	// --xrot --yrot --zrot
int ScanOptYPos ( ccp arg );			// --ypos
int ScanOptXSS	( uint xyz, ccp arg );		// --xss --yss --zss
int ScanOptAScale ( ccp arg );			// --ascale
int ScanOptARotate ( ccp arg );			// --arot

//-----------------------------------------------------------------------------

extern bool tform_script_enabled;

int ScanOptTformScript ( ccp arg );		// --tfor-script
bool ScanTformBegin();
bool TformScriptEnd();

enumError TformScriptCallF
(
    uint	dimension,	// '2' for a 2D and '3' for a 3D vector
    float3	*v,		// pointer to list of 3D vectors
    int		n,		// number of 3D vectors
    uint	off		// if n>1: offset from one to next vector
);

enumError TformScriptCallD
(
    uint	dimension,	// '2' for a 2D and '3' for a 3D vector
    double3	*v,		// pointer to list of 3D vectors
    int		n,		// number of 3D vectors
    uint	off		// if n>1: offset from one to next vector
);

//-----------------------------------------------------------------------------

void SetupConst();				// inititalize constants
int ScanOptConst ( ccp arg );			// --const
int ScanOptSetFlags ( ccp arg );		// --set-flags
int ScanOptSetScale ( ccp arg );		// --set-scale
int ScanOptSetRot ( ccp arg );			// --set-rot
int ScanOptSet ( uint xyz, ccp arg );		// --set

//-----------------------------------------------------------------------------

void TransformPosFloat2D
(
    float	*v,	// pointer to list of 2D vectors
    int		n,	// number of 2D vectors
    uint	off	// if n>1: offset from one to next vector
);

static inline void TransformPosFloat3D
(
    float	*v,	// pointer to list of 3D vectors
    int		n,	// number of 3D vectors
    uint	off	// if n>1: offset from one to next vector
)
{
    TransformF3NMatrixD(&opt_transform,(float3*)v,n,off);
    if (tform_script_enabled)
	TformScriptCallF(3,(float3*)v,n,off);
}

void TransformPosDouble2D
(
    double	*v,	// pointer to list of 2D vectors
    int		n,	// number of 2D vectors
    uint	off	// if n>1: offset from one to next vector
);

static inline void TransformPosDouble3D
(
    double	*v,	// pointer to list of 3D vectors
    int		n,	// number of 3D vectors
    uint	off	// if n>1: offset from one to next vector
)
{
    TransformD3NMatrixD(&opt_transform,(double3*)v,n,off);
    if (tform_script_enabled)
	TformScriptCallD(3,(double3*)v,n,off);
}

//-----------------------------------------------------------------------------

void TransformScaleFloat1D
(
    float	*v,	// pointer to list of 3D vectors
    int		n,	// number of 2D vectors
    uint	off	// if n>1: offset from one to next vector
);

void TransformScaleFloat3D
(
    float	*v,	// pointer to list of 3D vectors
    int		n,	// number of 2D vectors
    uint	off	// if n>1: offset from one to next vector
);

void TransformScaleDouble3D
(
    double	*v,	// pointer to list of 3D vectors
    int		n,	// number of 3D vectors
    uint	off	// if n>1: offset from one to next vector
);

//-----------------------------------------------------------------------------

void TransformRotFloat3D
(
    float	*v,	// pointer to list of 3D vectors
    int		n,	// number of 2D vectors
    uint	off	// if n>1: offset from one to next vector
);

void TransformRotDouble3D
(
    double	*v,	// pointer to list of 3D vectors
    int		n,	// number of 3D vectors
    uint	off	// if n>1: offset from one to next vector
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////		    scan text: Interface		///////////////
///////////////////////////////////////////////////////////////////////////////

ScanFile_t * AddSF
(
    ScanInfo_t		* si,		// base scan info
    ccp			data,		// data to scan
    uint		data_size,	// size of data
    ccp			name,		// NULL or name for warnings
    int			revision,	// revision number
    VarMap_t		* varmap	// NULL or new varmap (content moved)
);

//-----------------------------------------------------------------------------

bool DropSF
(
    ScanInfo_t		* si		// data structure
);

///////////////////////////////////////////////////////////////////////////////

void InitializeSI
(
    ScanInfo_t		* si,		// data structure to setup
    ccp			data,		// data to scan
    uint		data_size,	// size of data
    ccp			name,		// NULL or name for warnings
    int			revision	// revision number
);

//-----------------------------------------------------------------------------

void ResetSI
(
    ScanInfo_t		* si		// data structure to setup
);

//-----------------------------------------------------------------------------

void ResetLocalVarsSI
(
    ScanInfo_t		* si,		// data structure to setup
    int			revision	// revision number
);

//-----------------------------------------------------------------------------

void RestartSI
(
    ScanInfo_t		* si		// data structure to setup
);

//-----------------------------------------------------------------------------

char NextCharSI
(
    ScanInfo_t		* si,		// valid data
    bool		skip_lines	// true: skip empty lines
);

//-----------------------------------------------------------------------------

char SkipCharSI
(
    // return next non skipped character

    ScanInfo_t		* si,		// valid data
    char		ch		// character to skip
);

//-----------------------------------------------------------------------------

char NextLineSI
(
    ScanInfo_t		* si,		// valid data
    bool		skip_cont,	// true: skip continuation lines
    bool		allow_tcmd	// true: scan and execute text commands
);

//-----------------------------------------------------------------------------

enumError CheckLevelSI
(
    ScanInfo_t		* si		// valid data
);

//-----------------------------------------------------------------------------

void GotoEolSI
(
    ScanInfo_t		* si		// valid data
);

//-----------------------------------------------------------------------------

enumError CheckEolSI
(
    ScanInfo_t		* si		// valid data
);

//-----------------------------------------------------------------------------

enumError CheckWarnSI
(
    ScanInfo_t		* si,		// valid data
    char		ch,		// char to test
    enumError		err		// previous error level
);

//-----------------------------------------------------------------------------

enumError WarnIgnoreSI
(
    ScanInfo_t		* si,		// valid data
    ccp			format,		// NULL or format of text before 'line ignored'
    ...					// arguments for 'format'
)
__attribute__ ((__format__(__printf__,2,3)));

//-----------------------------------------------------------------------------

ccp FindLineFeedSI
(
    ScanInfo_t		* si,		// valid data
    ccp			ptr,		// start search here
    bool		mark_error	// true: increment error counter
);

//-----------------------------------------------------------------------------

ccp FindNextLineFeedSI
(
    ScanInfo_t		* si,		// valid data
    bool		mark_error	// true: increment error counter
);

///////////////////////////////////////////////////////////////////////////////

uint ScanNameSI
(
    // return length of name

    ScanInfo_t		* si,		// valid data
    char		* name_buf,	// destination buffer
    uint		buf_size,	// size of 'name_buf'
    bool		allow_minus,	// true: allow '-' in names
    bool		to_upper,	// true: convert letters to upper case
    int			* xyz		// not null: store vector index
					// -1: no index, 0=x, 1=y, 2=z
);

//-----------------------------------------------------------------------------

enumError ScanStringSI
(
    ScanInfo_t		* si,		// valid data
    char		** dest_str	// store string here, use FREE + MALLOC
);

//-----------------------------------------------------------------------------

extern char last_index_name[VARNAME_SIZE+1];

int ScanIndexSI
(
    ScanInfo_t		* si,		// valid data
    enumError		* err,		// return value: error code
    int			index,		// index to define if name
    IndexList_t		* il,		// NULL or index list
    int			insert_mode	// 0: don't insert
					// 1: insert into global namespace
					// 2: '1' + print "already defined" error
);

//-----------------------------------------------------------------------------

void SetIndexName ( ccp name );

//-----------------------------------------------------------------------------

const Var_t * FindVarSI
(
    ScanInfo_t		* si,		// not NULL: use var maps
    ccp			varname,	// variable name
    bool		print_warning	// true && 'si': print 'not found' warning
);

//-----------------------------------------------------------------------------

Var_t * OverwriteVarSI
(
    // returns always not NULL

    ScanInfo_t		* si,		// not NULL: use var maps for search
    VarMap_t		* vm_insert,	// valid variable map, insert search failed
    ccp			varname,	// varname to insert
    int			* xyz,		// not NULL: enable .X/Y/Z support
					//   input: <0: not a X/Y/Z
					//        0..2: is a X/Y/Z
					//	    >2: unknown
					//   output: store corrected value
    bool		* found		// not null: store found status (false=NEW)
);

//-----------------------------------------------------------------------------

bool DefineIndexInt
(
    // returns true, if inserted

    ScanInfo_t		* si,		// valid data
    ccp			suffix,		// name suffix
    int			value		// value to assign
);

//-----------------------------------------------------------------------------

bool DefineIndexDouble
(
    // returns true, if inserted

    ScanInfo_t		* si,		// valid data
    ccp			suffix,		// name suffix
    double		value		// value to assign
);

//-----------------------------------------------------------------------------

enumError ScanExprSI
(
    ScanInfo_t		* si,		// valid data
    Var_t		* var		// result: store value here
);

//-----------------------------------------------------------------------------

enumError ScanVectorExprSI
(
    // scans: v | a,b | x,y,z

    ScanInfo_t		* si,		// valid data
    Var_t		* var,		// result: store value here
    double		ab_default,	// used for 'a,b' format
    uint		xyz		// for ab default: 0:yz, 1:xz, 2:xy
);

enumError ScanVectorExpr
(
    // scans: v | x,z | x,y,z

    ccp			arg,		// string to scan
    ccp			name,		// name for error messages
    Var_t		* var		// result: store value here
);

//-----------------------------------------------------------------------------

enumError ScanMinMaxExprSI
(
    // scans: val (min=-val,max=+val) | val1,val2

    ScanInfo_t		*si,		// valid data
    Var_t		*min,		// result: store min vector here
    Var_t		*max		// result: store max vector here
);

//-----------------------------------------------------------------------------

enumError ScanValueSI
(
    ScanInfo_t		* si,		// valid data
    Var_t		* var		// result: store value here
);

//-----------------------------------------------------------------------------

enumError ScanUValueSI
(
    ScanInfo_t		* si,		// valid data
    long		* num,		// return value
    int			force_hex	// >0: force hex for non prefixed integers
);

//-----------------------------------------------------------------------------

enumError ScanDValueSI
(
    ScanInfo_t		* si,		// valid data
    double		* num		// return value
);

//-----------------------------------------------------------------------------

enumError ScanBitfieldSI
(
    ScanInfo_t		* si,		// valid data
    Var_t		* var		// result: store value here
);

//-----------------------------------------------------------------------------

enumError ScanU8SI
(
    ScanInfo_t		* si,		// valid data
    u8			* dest,		// destination list
    int			n,		// number of elements in 'dest'
    int			force_hex	// >0: force hex for non prefixed integers
);

//-----------------------------------------------------------------------------

enumError ScanU16SI
(
    ScanInfo_t		* si,		// valid data
    u16			* dest,		// destination list
    int			n,		// number of elements in 'dest'
    int			force_hex	// >0: force hex for non prefixed integers
);

//-----------------------------------------------------------------------------

enumError ScanU32SI
(
    ScanInfo_t		* si,		// valid data
    u32			* dest,		// destination list
    int			n,		// number of elements in 'dest'
    int			force_hex	// >0: force hex for non prefixed integers
);

//-----------------------------------------------------------------------------

enumError ScanBE16SI
(
    ScanInfo_t		* si,		// valid data
    u16			* dest,		// destination list
    int			n,		// number of elements in 'dest'
    int			force_hex	// >0: force hex for non prefixed integers
);

//-----------------------------------------------------------------------------

enumError ScanBE32SI
(
    ScanInfo_t		* si,		// valid data
    u32			* dest,		// destination list
    int			n,		// number of elements in 'dest'
    int			force_hex	// >0: force hex for non prefixed integers
);

//-----------------------------------------------------------------------------

enumError ScanBE32SI_swap
(
    ScanInfo_t		* si,		// valid data
    u32			* dest,		// destination list
    int			n,		// number of elements in 'dest'
    int			force_hex	// >0: force hex for non prefixed integers
);

//-----------------------------------------------------------------------------

enumError ScanFloatSI
(
    ScanInfo_t		* si,		// valid data
    float		* dest,		// destination list
    int			n		// number of elements in 'dest'
);

//-----------------------------------------------------------------------------

enumError ScanFloatV2SI
(
    ScanInfo_t		* si,		// valid data
    float		* dest,		// destination list
    int			n		// number of 2d vectors in 'dest'
);

//-----------------------------------------------------------------------------

enumError ScanFloatV3SI
(
    ScanInfo_t		* si,		// valid data
    float		* dest,		// destination list
    int			n		// number of 3d vectors in 'dest'
);

//-----------------------------------------------------------------------------

enumError ScanDoubleSI
(
    ScanInfo_t		* si,		// valid data
    double		* dest,		// destination list
    int			n		// number of elements in 'dest'
);

//-----------------------------------------------------------------------------

enumError ScanDoubleV3SI
(
    ScanInfo_t		* si,		// valid data
    double		* dest,		// destination list
    int			n		// number of elements in 'dest'
);

//-----------------------------------------------------------------------------

enumError ScanAssignIntSI
(
    ScanInfo_t		* si,		// valid data
    int			* value		// result: scanned value
);

//-----------------------------------------------------------------------------

enumError ScanAssignDoubleSI
(
    ScanInfo_t		* si,		// valid data
    double		* value		// result: scanned value
);

//-----------------------------------------------------------------------------

enumError ScanParamSI
(
    ScanInfo_t		* si,		// valid data
    ScanParam_t		* sp		// scan parameter
					// list ends with sp->name==NULL
);

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

TextCommand_t GetTextCommand
(
    ccp			name		// name to scan, upper case is assumed
);

//-----------------------------------------------------------------------------

enumError ScanTextCommand
(
    ScanInfo_t		* si		// valid data
);

//-----------------------------------------------------------------------------

enumError ExecTextCommand
(
    ScanInfo_t		* si,		// valid data
    TextCommand_t	tcmd		// command to execute
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			struct Line_t			///////////////
///////////////////////////////////////////////////////////////////////////////

#define MAX_LINE_HELPER 2
#define MAX_LINE_DIMEN  3

enum LineMode { LM_LINE, LM_BORDER, LM_BEZIER };

typedef struct Line_t
{
    uint	n;		// number of line points
    uint	n_pt;		// number of helper points: 0..2
    double	pt[MAX_LINE_HELPER][MAX_LINE_DIMEN]; // helper points
    int		line_mode;	// user defined mode

} Line_t;

//-----------------------------------------------------------------------------

void InitializeLine
(
    Line_t	* lin		// valid line data
);

enumError ScanLine
(
    Line_t	* lin,		// valid line data
    ScanInfo_t	* si,		// scanning source
    uint	n_dim,		// number of elements in each vector (<=3)
    int		mode		// user defined mode, stored in 'lin->mode'
);

void CalcLine
(
    Line_t	* lin,		// valid line data
    float	* dest,		// destination vector
    const float	* p1,		// source vector P1
    const float	* p2,		// source vector P2
    uint	n_dim,		// number of elements in each vector, n>3 possible
    uint	step		// value between 0 and lin->n inclcusive
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			matrix support			///////////////
///////////////////////////////////////////////////////////////////////////////

enumError cmd_matrix();

bool PrintTestMatrix( bool print_sep );

bool TestMatrix
(
    // returns false on error

    double3	*val,		// NULL or value to test
    uint	print_mode,	// 0: silent
				// 1: print vectors on error
				// 2: print vectors always
    bool	print_matrix	// true && if vector shall print:
				//   print matrix before, but only the first time
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			operators			///////////////
///////////////////////////////////////////////////////////////////////////////

typedef enum OpId_t
{
	OPI_NONE = 0,

	OPI_ADD,
	OPI_AND,
	OPI_DIV,
	OPI_EOR,
	OPI_EQ,
	OPI_EQEQ,
	OPI_GE,
	OPI_GT,
	OPI_LAND,
	OPI_LE,
	OPI_LEOR,
	OPI_LOR,
	OPI_LSHIFT,
	OPI_LT,
	OPI_MOD,
	OPI_MULT,
	OPI_NEQ,
	OPI_NEQNEQ,
	OPI_OR,
	OPI_POW,
	OPI_RSHIFT,
	OPI_SUB,

	OPI__N // = 23

} OpId_t;

//-----------------------------------------------------------------------------

void CalcExpr
(
    Var_t		* dst,		// first source and destination
    const Var_t		* src,		// second source
    OpId_t		op_id,		// operator id
    ScanInfo_t		* si		// not NULL: Use for warnings
);

//
///////////////////////////////////////////////////////////////////////////////
///////////////			    END				///////////////
///////////////////////////////////////////////////////////////////////////////

#endif // SZS_LIB_NUMERIC_H

