
#
###############################################################################
# Macros

@macro itembox
    #---------------------------------------------------------------
    # This macro helps to place N itemboxes in a row between
    # 2 points with optimal height.
    #---------------------------------------------------------------
    # Syntax:
    #	@:ITEMBOX(fall,1,pos)
    #	@:ITEMBOX(fall,n,pos1,pos2)
    #	@:ITEMBOX(fall,n,pos1,pos2,factor)
    #---------------------------------------------------------------
    # Parameters:
    #	FALL	: If >0: Raise the box by FALL and let it fall too
    #		: the road to get the optimal height.
    #	N	: If >0: Create N itemboxes between POS1 and POS2.
    #	pos1	: Position of first itembox.
    #	pos2	: Position of second itembox.
    #	factor	: If set and >1: Create a line from POS1 to POS2.
    #		: Enlarge the line by FACTOR in both directions.
    #		: Then use the outer points as new POS and POS2.
    #---------------------------------------------------------------
    # Examples:
    #	@:itembox( 200, 1, point )
    #	@:itembox(   0, 3, point1, point2 )
    #	@:itembox( 200, 4, v(2000,1000,3000), v(3000,1000,5000) )
    #---------------------------------------------------------------

    @param fall n p1 p2 factor
    @pdef fall = fall > 0 ? fall : 0
    @if isScalar(factor) > 1
	@pdef a = p1, b = p2, f = (1.0-factor)/2
	@pdef p1 = pos(f,a,b)
	@pdef p2 = pos(f,b,a)
    @endif
    @if n == 1
	@pdef p = fall ? kcl$fall(p1+vy(fall),200,KCL$DRIVE) : p1
	0 o$itembox p 0 0 0 0 -1  0 v3(0) 0 0 0 0 0x3f  v3(1)
    @else
	@pdef n = n - 1.0
	@for i=0;n
	    @pdef p = pos(i/n,p1,p2)
	    @doif(fall) @pdef p = kcl$fall(p+vy(fall),200,KCL$DRIVE)
	    0 o$itembox p 0 0 0 0 -1  0 v3(0) 0 0 0 0 0x3f  v3(1)
	@endfor
    @endif
@endmacro
